/*
 * File: ckpt_util.c
 *
 * Abstract: Various utilities such as error 
 *  logging and machine architecture/process 
 *  info etc.
 * 
 * (C) University of Amsterdam 
 *     Section Computational Science 
 *     Kruislaan 403, the Netherlands
 * 
 * CREATION DATE: unknown 
 * AUTHOR       : unknown
 * 
 * MODIFICATION HISTORY:
 * 
 * Date        By    Description 
 * ----------  ----  -----------------------------------
 * 08-09-1999  ZWH   Log ctime instead of plain time in CKPTLOG
 * 06-11-2003  TG    Adding new Dynamite ckpter 2.0 support
 */
#include <sys/types.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#ifndef __linux__
#include <stropts.h>
#endif
#include "ckpt.h"

#include <pvm3.h>
#include "tdpro.h"
#include "global.h"
#include "dpvmError.h"

int CkptWanted=0;
int InCriticalSection=0;
int setsigharderr;
int pvmmigrating = 0;

#if defined(DPVM) && defined(LOG_STDERR)
int logfd = -1;
#endif

#if 1
void dpvm_toggledebug();
#endif

extern char pvmckptdir[];
extern char pvmckptfile[];

extern int pvmmytid;

#ifdef DPVM
extern int mxfer_pvmmctl_scan_mode;
#endif

/* TG+ */
/*
 * TG: The USR1 signal handler for catching USR1 signal and
 * invoking the proper function from the ckpt library
 */

int (*dynamite_ckpt_ptr)(int, void *, void *);
void dpvm_usersave(char*);
void dpvm_userrestore(void);

void ckpt_handler(int signum) {
  int retval, ckptFN;
  char ckpt_file[PATH_MAX+1], ckptFN_file[PATH_MAX+1];

  dpvm_usersave(ckpt_file);

  sprintf(ckptFN_file, "%s%s\0", pvmckptdir, "/ckptFileName");
  ckptFN = open(ckptFN_file, O_CREAT | O_WRONLY, 0666);
  write(ckptFN, ckpt_file, strlen(ckpt_file));
  close(ckptFN);


  /* let's kill the process after checkpoint (2nd value 0) */
  retval = dynamite_ckpt_ptr(0, (void *)0, ckpt_file);

  dpvm_userrestore();
}
/* TG- */

/*
 * Print a message to the CKPTLOG file
 *
 * Four severity levels are recognized:
 * ERR_NORMAL, ERR_DEBUG, ERR_FATAL and ERR_PERR,
 * see also ckpt.h
 *
 * If a fatal error is encountered, the subroutine exits
 */
void ckpt_error(int err_type, char *fmt, ...)
{
  time_t currentTime;
  char timeString[25];
  va_list args;
  FILE *fp;
  struct utsname uts;

#ifndef DEBUG
  if (err_type & ERR_DEBUG)
    return;
#endif

  /* Print time, node and process ID */
  uname(&uts);
  fp = fopen("CKPTLOG", "a");
  if (!fp)
      return;
  currentTime = time(NULL);
  strcpy(timeString, ctime(&currentTime));
  timeString[24] = 0;
  fprintf(fp, "%s %-8s %5d |", timeString, uts.nodename, getpid());

  /* ERR_DEBUG */
  if (err_type & ERR_DEBUG)
    fprintf(fp, "DEBUG: ");

  /* ERR_NORMAL */
  va_start(args, fmt);
  vfprintf(fp, fmt, args);
  va_end(args);

  /* ERR_PERR */
  if (err_type & ERR_PERR)
      fprintf(fp, ": %s", strerror(errno));

  fprintf(fp, "\n");
  fflush(fp);

  fclose(fp);
  
  /* ERR_FATAL */
  if (err_type & ERR_FATAL)
      exit(1);
}

/* 
 * Return machine architecture
 *
 * Possible return values:
 * MACHINE_SUN4, MACHINE_SUN4M, MACHINE_SUN4C, 
 * MACHINE_I486, MACHINE_RS6000
 * 
 * Subroutine generates fatal error if machine type is
 * unknown or if 'uname' fails
 */
int get_mach()
{
  struct utsname uts;

  if (uname(&uts) < 0)
      ckpt_error(ERR_FATAL|ERR_PERR, "uname() failed");

  if (!strcmp(uts.machine, "sun4"))  return MACHINE_SUN4;
  if (!strcmp(uts.machine, "sun4m")) return MACHINE_SUN4M;
  if (!strcmp(uts.machine, "sun4c")) return MACHINE_SUN4C;
  if (!strcmp(uts.machine, "i486"))  return MACHINE_I486;

  if (!strcmp(uts.sysname, "AIX")) return MACHINE_RS6000;
  
  ckpt_error(ERR_FATAL, "unknown machine type '%s'", uts.machine);      

  return -1;
}

int Write(int fd, void *ptr, int nbytes)
{
  int nleft, nwritten;

  nleft = nbytes;

  while (nleft > 0) {
    nwritten = write(fd, ptr, nleft);
    if (nwritten <= 0)
      return nwritten; 		/* error */
    nleft -= nwritten;
    ptr    = &((char *)ptr)[nwritten];
  }
  return (nbytes - nleft);
}

int Read(int fd, void *ptr, int nbytes)
{
  int nleft, nread;

  nleft = nbytes;

  while (nleft > 0) {
    nread = read(fd, ptr, nleft);
    if (nread <= 0)
      return nread; 		/* error */
    else if (nread == 0)
      break;			/* EOF */

    nleft -= nread;
    ptr    = &((char *)ptr)[nread];
  }
  return (nbytes - nleft);
}

#ifdef TIME_CKPT
/*
 * Print elapsed time from <start_time> till now
 * to CKPTLOG file with <msg>
 */
void report_time(clock_t start_time, char *msg)
{
  clock_t end_time;
  struct tms after;

  end_time = times(&after);

  ckpt_error(ERR_NORMAL, "%s: %6.4fs", msg, (double)(end_time - start_time)/
#ifdef CLK_TCK
  CLK_TCK);
#else
  sysconf(3));
#endif
}
#endif

/*
 * int
 * fdsend(sock, fds, nfiles)
 * 	int sock, *fds, nfiles;
 * {
 * 	struct msghdr msg;
 * 	struct utsname uts;
 * 	int i;
 * 
 * 	uname(&uts);
 * 
 * 	if (!strcmp(uts.release, "5.5") || !strcmp(uts.release, "5.5.1")) {
 * 		for (i = 0; i < nfiles; i++) {
 * 			if (ioctl(sock, I_SENDFD, fds[i]) < 0) {
 * 				perror("ioctl I_SENDFD");
 * 				return -1;
 * 			}
 * 		}
 * 	} else {
 * 		memset(&msg, 0, sizeof(msg));
 * 		msg.msg_accrights = (caddr_t) fds;
 * 		msg.msg_accrightslen = nfiles * sizeof (int);
 * 
 * 		if (sendmsg(sock, &msg, 0)) {
 * 			perror("sendmsg");
 * 			return -1;
 *		}
 * 	}
 * 	return 0;
 * }
 */

void dpvm_usersave(char *name)
{
        char filename[PATH_MAX+1], *login;
	char hostname[128];
	int old2, log;
#ifdef DEBUG
        fflush(stderr);
	old2 = dup(2);

        strcpy(filename, "/tmp/dpvm_usersave.out.");
        login = getenv("LOGNAME");
	if (!login)
	    login = getenv("USER");
	strcat(filename, login);

	log = open(filename, O_WRONLY | O_APPEND | O_CREAT, 0666);
	if (log > 0)
	{
	    dup2(log, 2);
	    close(log);
	}

	dpvm_toggledebug(0);
#endif
	mxfer_pvmmctl_scan_mode = MXSM_NONE;
	ttpcb_flushall();
	pvmmigrating = 1;
	gethostname(hostname, sizeof hostname);
        sprintf(name, "%s/%s.%x.%x" /* Checkpoint filename is ...*/
          , pvmckptdir              /* ...pvm ckpt dir...*/
          , hostname                /* ... + hostname ...*/
          , pvmmytid                /* ... + pvm tid  ...*/
          , time(NULL));            /* ... + time stamp */
#ifdef DEBUG
	dpvm_toggledebug(0);

	fflush(stderr);
	dup2(old2, 2);
	close(old2);
#endif
	pvmendtask();
}

void dpvm_userrestore()
{
	extern int pvmmyptid;

	pvmbeatask();
	pvmmigrating = 0;
	mxfer_pvmmctl_scan_mode = MXSM_ALL;
#if defined(LOG_STDERR)
	if (pvmmyptid != 0) {
		ckpt_error(ERR_DEBUG, "dup2(%d, 2)\n", logfd);
		dup2(logfd, 2);
	}
#endif
	ckpt_error(ERR_DEBUG, "end of dpvm_userrestore\n");
}

extern int debugmask;
int oldmask, debugging = 0;

void
dpvm_toggledebug(sig)
	int sig;
{
#ifdef __linux__
    signal(SIGUSR2, dpvm_toggledebug);
#endif
	if (debugging) {
		debugmask = oldmask;
		debugging = 0;
		write(2, "debugging off\n", 14);
	} else {
		debugmask = 0xffffffff;
		debugging = 1;
		write(2, "debugging on\n", 13);
	}
}
