
/*
 * PVM version 3.3:  Parallel Virtual Machine System University of
 * Tennessee, Knoxville TN. Oak Ridge National Laboratory, Oak Ridge
 * TN. Emory University, Atlanta GA. Authors:  A. L. Beguelin, J. J.
 * Dongarra, G. A. Geist, W. C. Jiang, R. J. Manchek, B. K. Moore,
 * and V. S. Sunderam (C) 1992 All Rights Reserved
 * 
 * NOTICE
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 * 
 * Neither the Institutions (Emory University, Oak Ridge National
 * Laboratory, and University of Tennessee) nor the Authors make any
 * representations about the suitability of this software for any
 * purpose.  This software is provided ``as is'' without express or
 * implied warranty.
 * 
 * PVM version 3 was funded in part by the U.S. Department of Energy,
 * the National Science Foundation and the State of Tennessee.
 */

/*
 * lpvm.c
 * 
 * Libpvm core for unix environment.
 * 
 * $Log: lpvm.c,v $
 * Revision 1.25  2001/10/15 11:04:00  kamil
 * In pvm_tc_conack, when blocking for the SOC message, accepting new direct
 * 	connections is essential to avoid deadlocks.
 *
 * Revision 1.24  2000/11/15 13:07:33  kamil
 * Don't store out-of-sequence message in pcbp, create a global list oosrxlist
 * 	for them.
 *
 * Revision 1.23  2000/10/06 08:49:35  kamil
 * In out-of-sequence-message handling code, ignore not only messages from the
 * 	local pvmd, but also from remote pvmds.
 *
 * Revision 1.22  2000/05/10 08:04:54  kamil
 * Get rid of this weird "tvp_was_zero" stuff: mroute can now be called
 *    with timeout > 1.0.
 * Use bytes of data as packet sequence numbers, so that packets can be
 *    fragmented.
 * Only block signals in StartCriticalSection() when InCriticalSection is 0.
 *
 * Revision 1.21  2000/03/10 12:25:26  kamil
 * New tseq structure and accompanying infrastructure for separate pool of
 *    message sequence identifiers local for every src,dst task pair.
 * Fix a bug in fragment sequencing handling in mxinput.
 * Add message sequencing handling to mxinput and mxfer.
 *
 * Revision 1.20  2000/02/14 09:49:12  kamil
 * Add additional accept-connect synchronisation via the start-of-connection
 * 	message.
 * Mark the peer task as dead in dpvm_oobhandler even if in TTCONWAIT state.
 *
 * Revision 1.19  2000/02/04 12:44:50  kamil
 * mxfer_pvmmctl_scan_all changed to mxfer_pvmmctl_scan_mode.
 * Minor output changes in pvmlogerror timestamp, timestamp added to pvmlogperror.
 * Changed protocol of closing direct connections: now migrating task sends
 *    a TC_EOC first, the peer replies with TC_EOC and waits for EOF.
 * When connect() fails in pvm_tc_conack, mark connection as DEAD, not DENIED.
 * Merge the code to handle CKPTSIGNAL and SIGURG in mxfer.
 * Handle pvmmctl_scan_mode == MXSM_NONE.
 *
 * Revision 1.18  2000/02/02 12:06:55  kamil
 * Remove the wantmore_ptr hack: it was incomplete and difficult to do right.
 * 	Instead, zero wantmore in mxfer after SIGURG/CKPT signals.
 * Add mxfer_pvmmctl_scan_all, to rescan the whole input list in mxfer during
 * 	the first iteration or after SIGURG/CKPT signals.
 * Comment out the tc_moved_handled_by_pvmmctl hack: it is probably not needed
 * 	after all.
 * Wait for TC_EOC from topvmd connection before migrating.
 * More verbose DEBUG output in ttpcb_flushall.
 * Better handling of not fully established direct connections in ttpcb_flushall.
 * When establishing a direct connection in mroute, abort if ttocb is no longer
 * 	available (after the migration).
 * Disable non-blocking direct connections.
 * In pvm_tc_conreq, properly handle second CONREQ if already in TTGRNWAIT state.
 * Make the direct connection fully symetrical: SO_LINGER was missing on the
 * 	non-req side.
 * In dpvm_oobhandler, don't mark a connection as dead if it is in TTCONWAIT
 * 	state.
 *
 * Revision 1.17  2000/01/14 10:00:09  kamil
 * Handle TC_MOVED in pvmmct() again.
 * Remove accidentally checked in debug message from dpvm_oobhandler.
 *
 * Revision 1.16  2000/01/11 11:48:12  kamil
 * New tt_migrating field in ttpcb, marks a connection that should not be
 * 	used for writing anymore.
 * Completelety revamp the protocol of closing direct connections.
 * In ttpcb_flushall, wait for TC_EOC messages from peer tasks.
 * Use PVM 3.4 code for handling crossed connection requests.
 * TC_MOVED no longer handled in pvmmctl().
 * Remove Linux hacks for handling CKPTSIGNAL, they turn out not to be
 * 	necessary.
 * Check for the pending SIGURG signal, too.
 * If mxinput fails, only attempt to resend a packet if it wasn't completely
 * 	sent.
 * New dpvm_oobhandler function for handling of the SIGURG signal.
 * Under Linux, use signal, not sigaction to set up signal handlers.
 * Mask/unmask SIGURG too in {Start,End}CriticalSection.
 *
 * Revision 1.15  1999/12/02  10:05:18  kamil
 * In pvmmctl(), case TC_MOVED, mark connection as dead, so as to prevent mxfer()
 * from sending any more data through it. Modify mxfer() accordingly, to check
 * the state of the connection and send the message via pvmd if connection is
 * not open.
 * In mxfer(), after migration don't use direct connections, resend the whole
 * message via pvmd.
 * In mxfer(), initialise sequence numbers for queued messages and those
 * generated in pvmmctl().
 *
 * Revision 1.14  1999/11/15  09:47:26  kamil
 * Include sys/ioctl.h under Linux, not filio.h.
 * Fix a compiler warning message.
 *
 * Revision 1.13  1999/11/09  14:38:36  zegerh
 * Initial Linux checkin.
 * Revision 1.12  1999/06/24  11:36:27  dpvm Add
 * explicit 'void' type to some functions. Check-in test.
 * 
 * Revision 1.11  1999/06/21  08:34:27  frank Make pvm_mcast atomic; if
 * we get moved to a new host it will not understand the multicast
 * address.
 * 
 * Forward t-d protocol messages correctly.
 * 
 * Don't wait for an answer for TM_MIG.
 * 
 * Merge queue of received messages for the empty contect in conn2.
 * 
 * Revision 1.10  1999/06/18 09:16:52  frank Deal with out-of-sequence
 * fragments. Make the daemons do forwarding. Check messages'
 * destination at the last moment with a routing table check. Solve a
 * racecondition in t-d communication when migrating.
 * 
 * Revision 1.9  1999/06/07 08:06:46  frank Remove possible deadlock
 * when checkpointing a task. The daemon will wait for the task to
 * flush its messages, but the task may be in a select without
 * timeout, waiting for the daemon.
 * 
 * Revision 1.8  1999/06/04 08:32:58  frank Fix multicast code. We can
 * not assume that the hostpart of a tid actually refers to the host
 * that it's running on, since it may have migrated.
 * 
 * Revision 1.7  1999/05/03 09:21:18  frank Conditionalize stderr
 * logging on LOG_STDERR
 * 
 * Revision 1.6  1999/05/03 09:15:14  frank Implement better checkpoint
 * naming. Use DPVM_CKPTDIR environment variable for checkpoint
 * storage.
 * 
 * Revision 1.5  1999/04/29 12:10:05  frank Extend the checkpointing
 * protocol to include the name of the checkpoint file.
 * 
 * Revision 1.4  1999/04/27 09:50:19  frank handle direct connections
 * don't re-initialize sockets when restarting, and certainly do not
 * throw the buffers away
 * 
 * Revision 1.3  1999/04/15 08:28:23  frank Checkin of half-baked direct
 * connection support, to set a mark for the latest version exported
 * to Paderborn.
 * 
 * Revision 1.2  1999/02/02 17:02:35  frank ERR_PERROR -> ERR_PERR
 * 
 * Use CKPTFILE for checkpoint filename.
 * 
 * Revision 1.1.1.1  1998/09/25  12:43:45  frank First Dynamite import
 * 
 * Revision 1.15  1995/02/01  21:08:46  manchek error 4 is now
 * PvmOverflow. connect is retried in case of EINTR
 * 
 * Revision 1.14  1994/12/20  16:28:19  manchek added PvmShowTids case
 * to setopt. removed EOF message on pvmd socket close
 * 
 * Revision 1.13  1994/10/15  19:07:31  manchek don't use fd_sets when
 * select returns -1; will hang reading. cast message tags for
 * comparison as integers
 * 
 * Revision 1.12  1994/09/02  15:23:37  manchek fixed segfaults in
 * setopt when task not connected
 * 
 * Revision 1.11  1994/07/18  19:20:18  manchek fix to call write() with
 * max 4096 length for RS6K
 * 
 * Revision 1.10  1994/06/21  19:36:23  manchek forgot to ifdef tt_spath
 * 
 * Revision 1.9  1994/06/21  18:31:31  manchek connect in mksocs() tries
 * several times before giving up. don't setsockopt() at TCP level on
 * Unix task-task sockets
 * 
 * Revision 1.8  1994/06/04  21:44:57  manchek added unix domain sockets
 * 
 * Revision 1.7  1994/06/03  20:38:15  manchek version 3.3.0
 * 
 * Revision 1.6  1993/11/30  23:50:01  manchek set nodelay and snd, rcv
 * buffer sizes on t-t sockets
 * 
 * Revision 1.5  1993/11/30  15:51:30  manchek beatask complains if it
 * can't write d-auth file (fs full?)
 * 
 * Revision 1.4  1993/10/04  20:29:05  manchek mksocks() and
 * pvm_start_pvmd() now use pvmdsockfile(), not TDSOCKNAME. made
 * pvm_useruid global for pvmcruft.c
 * 
 * Revision 1.3  1993/10/04  19:10:22  manchek mctl() conflicts with
 * mctl(2) - declare static for now
 * 
 * Revision 1.2  1993/09/16  21:36:20  manchek pvm_start_pvmd() was
 * freeing the return string from pvmgetpvmd()
 * 
 * Revision 1.1  1993/08/30  23:26:48  manchek Initial revision
 * 
 */


#include <stdio.h>
#ifdef IMA_BSD386
#include <machine/endian.h>
#endif
#ifdef IMA_LINUX
#include <endian.h>
#endif
#include <rpc/types.h>
#include <rpc/xdr.h>
#include <pvm3.h>
#include <sys/socket.h>
#ifndef NOUNIXDOM
#include <sys/un.h>
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#if defined(IMA_RS6K) || defined(IMA_POWER4)
#include <sys/select.h>
#endif
#include <sys/stat.h>
#include <fcntl.h>
#ifdef	SYSVSTR
#include <string.h>
#define	CINDEX(s,c)	strchr(s,c)
#else
#include <strings.h>
#define	CINDEX(s,c)	index(s,c)
#endif
#include <errno.h>
#include <signal.h>
#include "global.h"
#include "tdpro.h"
#include "ddpro.h"
#include "pvmalloc.h"
#include "pvmfrag.h"
#include "pvmumbuf.h"
#include "listmac.h"
#include "tvdefs.h"
#include "bfunc.h"
#include <pvmtev.h>
#include "tevmac.h"

#include <pwd.h>

#ifdef DPVM
#ifdef __sun__
#include <sys/filio.h>
#elif __linux
#include <sys/ioctl.h>
#endif
#include "ckpt.h"
#include "monlib.h"
#endif	/* DPVM */
#include "error.h"

#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif

#ifndef	min
#define	min(a,b)	((a)<(b)?(a):(b))
#endif

/*
 * task-task or task-pvmd control block
 */

struct ttpcb
{
  struct ttpcb   *tt_link;	/* dll of peers */
  struct ttpcb   *tt_rlink;
  int             tt_tid;	/* tid of peer or 0 if master of list */
  int             tt_state;
  int             tt_fd;	/* fd of connection or -1 */
  struct sockaddr_in tt_sad;	/* address of our socket */
  struct sockaddr_in tt_osad;	/* (authd) address of peer socket */
  struct umbuf   *tt_rxfrag;	/* not-assembled incm msg */
  struct frag    *tt_rxf;	/* partial incm frag */
#ifndef NOUNIXDOM
  char           *tt_spath;	/* socket path */
#endif
#ifdef DPVM
  char            tt_migrating; /* peer is about to migrate: we could use
				   tt_state, but too many places depend on
				   TTOPEN, so it's easier this way. */
#endif
};

/* ttpcb states */
#define	TTCONWAIT	1	/* requested, waiting for reply to
				 * connect */
#define	TTGRNWAIT	2	/* granted, waiting for connect */
#define	TTOPEN		3	/* connection running */
#define	TTDENY		4	/* don't attempt connection */
#define	TTDEAD		5	/* task dead; pcb should be cleared */

#ifdef DPVM
#define TT_SOC		0xff	/* start-of-connection message for direct
				   connections */
#endif

#ifdef DPVM
/* List of message sequence numbers there is a separate pool for every
   task, this list stores last id sent and received.  */

struct tseq
{
  struct tseq *ts_link;
  struct tseq *ts_rlink;
  int          ts_tid;
  int          ts_lastsend;
  int          ts_lastrecv;
};
#endif

#ifndef	TTSOCKBUF
#define	TTSOCKBUF	0x8000
#endif

#ifdef	NOTMPNAM
#define	TMPNAMFUN(x)	pvmtmpnam(x)
#define	LEN_OF_TMP_NAM	32
char           *pvmtmpnam();

#else				/* NOTMPNAM */
#define	TMPNAMFUN(x)	tmpnam(x)
#ifdef	L_tmpnam
#define	LEN_OF_TMP_NAM	L_tmpnam
#else
#define	LEN_OF_TMP_NAM	64
#endif
#endif				/* NOTMPNAM */

char           *getenv();

extern struct encvec *enctovec();
extern char    *inadport_decimal();
extern char    *inadport_hex();
char           *pvmgetpvmd();
char           *pvmdsockfile();

void            dpvm_set();


/***************
 **  Globals  **
 **           **
 ***************/

extern int      errno;		/* from libc */
#ifndef __linux__
extern char    *sys_errlist[];
#endif __linux__
//extern int      sys_nerr;

extern int      pvmrbufmid;	/* from pack.c */
extern int      pvmsbufmid;	/* from pack.c */

extern int      pvmmigrating;

char            pvmckptfile[MAXPATHLEN];
char            pvmckptdir[MAXPATHLEN];

char           *pvm_errlist[] = {	/* error messages for
					 * -pvm_errno */
  "Error 0",
  "Error 1",
  "Bad parameter",
  "Count mismatch",
  "Value too large",
  "End of buffer",
  "No such host",
  "No such file",
  "Error 8",			/* not used */
  "Error 9",			/* not used */
  "Malloc failed",
  "Error 11",			/* not used */
  "Can't decode message",
  "Error 13",			/* not used */
  "Can't contact local daemon",
  "No current buffer",
  "No such buffer",
  "Null group name",
  "Already in group",
  "No such group",
  "Not in group",
  "No such instance",
  "Host failed",
  "No parent task",
  "Not implemented",
  "Pvmd system error",
  "Version mismatch",
  "Out of resources",
  "Duplicate host",
  "Can't start pvmd",
  "Already in progress",
  "No such task",
  "No such entry",
  "Duplicate entry",
};

int             pvmautoerr = 1;	/* whether to auto print err msgs */
int             pvmcouttid = 0;	/* child stdout dst and code */
int             pvmcoutcod = 0;
int             pvmctrctid = 0;	/* child trace dst and code */
int             pvmctrccod = 0;
int             pvmfrgsiz = UDPMAXLEN;	/* message frag length (to
					 * pack) */
int             pvmmyptid = -1;	/* parent task id */
int             pvmmytid = -1;	/* this task id */
int             pvmmyupid = -1;	/* process unix pid */
int             pvmudpmtu = UDPMAXLEN;	/* local UDP MTU */
int             pvm_errno = 0;	/* last libpvm error code */
int             pvmmyndf = 0;	/* host native data enc, init XDR */
int             pvm_nerr = sizeof(pvm_errlist)
/ sizeof(pvm_errlist[0]);	/* exported num of errors */
int             pvmschedtid = 0;/* scheduler task */
int             pvmrescode = 0;	/* allow use of reserved tids, codes */
struct umbuf   *pvmrxlist = 0;	/* not-recvd msg list */
int             pvmtidhmask = TIDHOST;	/* mask for host field of
					 * tids */
int             pvmouttid = 0;	/* stdout dst and code */
int             pvmoutcod = 0;
int             pvmshowtaskid = 1;	/* catchout shows task ids */
int             pvmtoplvl = 1;	/* function called from outside lib */
int             pvmtrctid = 0;	/* trace dst and code */
int             pvmtrccod = 0;
Pvmtmask        pvmtrcmask;	/* trace bitmask */
Pvmtmask        pvmctrcmask;	/* child trace bitmask */
int             pvm_useruid = -1;	/* user's unix uid */

#ifdef DPVM
int             otid = -1;	/* original tid */
int             optid = -1;	/* original parent tid */
int             mxfer_pvmmctl_scan_mode = MXSM_ALL;
struct tseq    *tslist = 0;   /* list of task sequence numbers */
extern int      CkptWanted;
extern int      InCriticalSection;
struct umbuf   *oosrxlist = 0;	/* out-of-sequence msg list */

int tseq_get_newsend(int tid);
int tseq_get_lastrecv(int tid);
void tseq_increase_lastrecv(int tid);
#endif


/***************
 **  Private  **
 **           **
 ***************/

static char     rcsid[] = "$Id: lpvm.c,v 1.25 2001/10/15 11:04:00 kamil Exp $";
int             debugmask = 0;	/* which debugging info */
static int      mxfersingle = 1;/* mxfer returns after single frag */
static char     pvmtxt[512];	/* scratch for error log */
static struct sockaddr_in pvmourinet;	/* our host ip addr */

#ifdef DPVM
/* static int      tc_moved_handled_by_pvmmctl = 0; */
#ifdef NO_DIRECT
static int      pvmrouteopt = PvmDontRoute;	/* task-task routing
						 * style */

#else
static int      pvmrouteopt = PvmAllowDirect;	/* task-task routing
						 * style */

#endif
#else				/* tsja.. */
static int      pvmrouteopt = PvmRouteDirect;	/* task-task routing
						 * style */

#endif
static struct ttpcb *ttlist = 0;/* dll of connected tasks */
static struct ttpcb *topvmd = 0;/* default route (to pvmd) */
int             pvmnfds = 0;	/* 1 + highest bit set in fds */
fd_set          pvmrfds;	/* rd fdset for mxfer() */
static int      pvmtrcmid = 0;	/* message buffer for trace */
static int      pvmtrcsbf = 0;	/* existing other buffer during trace */


/**************************
 **  Internal Functions  **
 **                      **
 **************************/

/*
 * pvmbailout()
 * 
 * Called by low-level stuff in f.e. pvmfrag.c.  Don't really want to
 * bail in libpvm.
 */

void
pvmbailout(n)
  int             n;
{
  n = n;			/* ayn rand was here */
}


/*
 * pvmlogerror()
 * 
 * Log a libpvm error message.  Prepends a string identifying the task.
 */

void
pvmlogerror(s)
  char           *s;
{
  char            hname[128];
  time_t          t;
  struct tm      *tm;

  gethostname(hname, sizeof hname);
  time(&t);
  tm = localtime(&t);

  if (pvmmytid == -1)
    fprintf(stderr, "%s %02d:%02d:%02d libpvm [pid%d]: %s", hname, tm->tm_hour, tm->tm_min, tm->tm_sec, pvmmyupid, s);
  else
    fprintf(stderr, "%s %02d:%02d:%02d libpvm [t%x]: %s", hname, tm->tm_hour, tm->tm_min, tm->tm_sec, pvmmytid, s);

  fflush(stderr);
}


/*
 * pvmlogperror()
 * 
 * Log a libpvm error message.  Prepends a string identifying the task
 * and appends the system error string for _errno.
 */

void
pvmlogperror(s)
  char           *s;
{
  const char     *em;
  char            hname[128];
  time_t          t;
  struct tm      *tm;

  gethostname(hname, sizeof hname);
  time(&t);
  tm = localtime(&t);

  em = ((errno >= 0/* && errno < sys_nerr*/)
	? strerror(errno) : "Unknown Error");
  if (pvmmytid == -1)
    fprintf(stderr, "%s %02d:%02d:%02d libpvm [pid%d]: %s: %s\n", hname, tm->tm_hour, tm->tm_min, tm->tm_sec, pvmmyupid, s, em);
  else
    fprintf(stderr, "%s %02d:%02d:%02d libpvm [t%x]: %s: %s\n", hname, tm->tm_hour, tm->tm_min, tm->tm_sec, pvmmytid, s, em);
}


pvm_fd_add(fd, sets)
  int             fd;		/* the fd */
  int             sets;		/* which sets */
{
#ifdef	SANITY
  if (fd < 0 || fd >= FD_SETSIZE)
  {
    sprintf(pvmtxt, "pvm_fd_add() bad fd %d\n", fd);
    pvmlogerror(pvmtxt);
    return 1;
  }
#endif
  if (sets & 1)
    FD_SET(fd, &pvmrfds);
  /*
   * if (sets & 2) FD_SET(fd, &pvmwfds); if (sets & 4) FD_SET(fd,
   * &pvmefds);
   */

  /* if this is new highest, adjust nfds */

  if (fd >= pvmnfds)
    pvmnfds = fd + 1;
  return 0;
}


pvm_fd_delete(fd, sets)
  int             fd;		/* the fd */
  int             sets;		/* which sets */
{
#ifdef	SANITY
  if (fd < 0 || fd >= FD_SETSIZE)
  {
    sprintf(pvmtxt, "pvm_fd_delete() bad fd %d\n", fd);
    pvmlogerror(pvmtxt);
    return 1;
  }
#endif
  if (sets & 1)
    FD_CLR(fd, &pvmrfds);
  /*
   * if (sets & 2) FD_CLR(fd, &pvmwfds); if (sets & 4) FD_CLR(fd,
   * &pvmefds);
   */

  /* if this was highest, may have to adjust nfds to new highest */

  if (fd + 1 == pvmnfds)
    while (pvmnfds > 0)
    {
      pvmnfds--;
      if (FD_ISSET(pvmnfds, &pvmrfds)
      /*
       * || FD_ISSET(pvmnfds, &pvmefds) || FD_ISSET(pvmnfds,
       * &pvmwfds)
       */
	)
      {
	pvmnfds++;
	break;
      }
    }
  return 0;
}


print_fdset(pad, n, f)
  char           *pad;		/* label at head */
  int             n;		/* max fd + 1 */
  fd_set         *f;		/* fd set */
{
  char           *p = pvmtxt;
  int             i;
  char           *s = "";

  strcpy(p, pad);
  p += strlen(p);
  for (i = 0; i < n; i++)
    if (FD_ISSET(i, f))
    {
      sprintf(p, "%s%d", s, i);
      p += strlen(p);
      s = ",";
    }
  strcat(p, "\n");
  pvmlogerror(pvmtxt);
  return 0;
}


/*
 * ttpcb_new()
 * 
 * Create a new, blank ttpcb.
 */

struct ttpcb   *
ttpcb_new()
{
  struct ttpcb   *pcbp;

  if (pcbp = TALLOC(1, struct ttpcb, "tpcb"))
  {
    BZERO((char *) pcbp, sizeof(struct ttpcb));
    pcbp->tt_fd = -1;
    pcbp->tt_rxfrag = TALLOC(1, struct umbuf, "umb");
    BZERO((char *) pcbp->tt_rxfrag, sizeof(struct umbuf));
    pcbp->tt_rxfrag->ub_link = pcbp->tt_rxfrag->ub_rlink = pcbp->tt_rxfrag;
#ifdef DPVM
    pcbp->tt_migrating = 0;
#endif
  }
  return pcbp;
}


/*
 * ttpcb_delete()
 * 
 * Get rid of a ttpcb.  Delete it from any list, close the socket, free
 * any rx frag heap.
 */

void
ttpcb_delete(pcbp)
  struct ttpcb   *pcbp;
{
  struct umbuf   *up;

  if (pcbp->tt_link)
  {
    LISTDELETE(pcbp, tt_link, tt_rlink);
  }
  if (pcbp->tt_fd != -1)
  {
    pvm_fd_delete(pcbp->tt_fd, 3);
    (void) close(pcbp->tt_fd);
  }
  if (up = pcbp->tt_rxfrag)
  {
    while (up->ub_link != up)
      umbuf_free(up->ub_link->ub_mid);
    PVM_FREE(up);
  }
  if (pcbp->tt_rxf)
    fr_unref(pcbp->tt_rxf);
#ifndef NOUNIXDOM
  if (pcbp->tt_spath)
    (void) unlink(pcbp->tt_spath);
#endif

  PVM_FREE(pcbp);
}


/*
 * ttpcb_creat()
 * 
 * Create a new task-task entry.  Allocates a ttpcb, inserts it in
 * ttlist.
 */

struct ttpcb   *
ttpcb_creat(tid)
  int             tid;		/* peer tid */
{
  struct ttpcb   *pcbp, *pcbp2;

  if (pcbp = ttpcb_new())
  {
    pcbp->tt_tid = tid;

    for (pcbp2 = ttlist->tt_link; pcbp2 != ttlist; pcbp2 = pcbp2->tt_link)
      if (pcbp2->tt_tid > tid)
	break;
    LISTPUTBEFORE(pcbp2, pcbp, tt_link, tt_rlink);
  }
  return pcbp;
}


/*
 * ttpcb_find()
 * 
 * Find a task-task entry by tid in ttlist.
 */

struct ttpcb   *
ttpcb_find(tid)
  int             tid;		/* peer tid */
{
  struct ttpcb   *pcbp;

  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
    if (pcbp->tt_tid >= tid)
      break;
  return (pcbp->tt_tid == tid) ? pcbp : (struct ttpcb *) 0;
}


/*
 * ttpcb_dead()
 * 
 * Mark ttpcb dead, close socket, remove it from fd sets.
 */

void
ttpcb_dead(pcbp)
  struct ttpcb   *pcbp;
{
  struct umbuf   *up;

  pcbp->tt_state = TTDEAD;
  if (pcbp->tt_fd != -1)
  {
    pvm_fd_delete(pcbp->tt_fd, 3);
    (void) close(pcbp->tt_fd);
    pcbp->tt_fd = -1;
  }
#ifndef NOUNIXDOM
  if (pcbp->tt_spath)
  {
    (void) unlink(pcbp->tt_spath);
    pcbp->tt_spath = 0;
  }
#endif
  if (pcbp->tt_rxf)
  {
    fr_unref(pcbp->tt_rxf);
    pcbp->tt_rxf = 0;
  }
  if (up = pcbp->tt_rxfrag)
  {
    while (up->ub_link != up)
      umbuf_free(up->ub_link->ub_mid);
  }
}


int
ttpcb_dump(pcbp)
  struct ttpcb   *pcbp;
{
  sprintf(pvmtxt, "ttpcb_dump() t%x fd=%d sad=%s",
	  pcbp->tt_tid,
	  pcbp->tt_fd,
	  inadport_decimal(&pcbp->tt_sad));
  sprintf(pvmtxt + strlen(pvmtxt), " osad=%s state=%d\n",
	  inadport_decimal(&pcbp->tt_osad), pcbp->tt_state);
  pvmlogerror(pvmtxt);
  return 0;
}


int
ttpcb_dumpall()
{
  struct ttpcb   *pcbp;

  pvmlogerror("ttpcb_dumpall()\n");
  ttpcb_dump(topvmd);
  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
    ttpcb_dump(pcbp);
  return 0;
}

int
ttpcb_getfds(fds, tids, ntids)
  int           **fds, **tids, *ntids;
{
  int             n = 0;
  struct ttpcb   *pcbp;

  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
    if (pcbp->tt_state == TTOPEN || pcbp->tt_state == TTGRNWAIT)
      n++;

  *tids = TALLOC(n, int, "passfd tids");
  *fds = TALLOC(n, int, "passfd fds");

  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
  {
    if (pcbp->tt_state == TTOPEN || pcbp->tt_state == TTGRNWAIT)
    {
      (*tids)[n] = pcbp->tt_tid;
      (*fds)[n] = pcbp->tt_fd;
      n++;
    }
  }

  *ntids = n;
  return 0;
}

/*
 * Make sure that there is no data left to be read on any sockets. Do
 * this by shutting down the socket for reading, calling mxinput
 * (thus receiving all data that was buffered) so that we receive
 * possibly complete messages. We may dump fragments, but that's ok,
 * since the other side will notice that and restart at the first
 * fragment.
 */
int ttpcb_flushall()
{
  struct ttpcb   *pcbp;
  int             sbf, rbf, dummy;
  int             n = 0, one = 1;
  int             tid = htonl(pvmmytid);
  int             cc;
  int             received_eoc_from_pvmd = 0;
  struct timeval ztv;

#if 0
  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
  {
    if (pcbp->tt_state == TTOPEN)
    {
      ioctl(pcbp->tt_fd, FIONREAD, &n);
      sprintf(pvmtxt, "%d bytes ready for pcb %x\n", n,
	      pcbp->tt_tid);
      pvmlogerror(pvmtxt);
      ioctl(pcbp->tt_fd, FIONBIO, &one);
      while (mxinput(pcbp, 1) != -1);
      if (shutdown(pcbp->tt_fd, 2) < 0)
	pvmlogperror("shutdown");
    }
  }
#else

  /* Begin by sending TC_EOC messages to all the peers with which we have an
     active direct connection open.  Also get rid of any partially-open
     direct connections.  */
  if (debugmask & TDMROUTE)
      pvmlogerror("ttpcb_flushall() direct connections list:\n");
  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
  {
      if (debugmask & TDMROUTE)
      {
	  sprintf(pvmtxt, "ttpcb_flushall() t%x state %d\n", pcbp->tt_tid,
		  pcbp->tt_state);
	  pvmlogerror(pvmtxt);
      }
      if (pcbp->tt_state == TTOPEN)
      {
	  if (debugmask & TDMROUTE)
	  {
	      sprintf(pvmtxt, "ttpcb_flushall() sending TC_EOC to t%x\n", pcbp->tt_tid);
	      pvmlogerror(pvmtxt);
	  }
	  sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	  dummy = 0;
	  pvm_pkint(&dummy, 1, 1);
	  ztv.tv_sec = ztv.tv_usec = 0;
	  cc = mroute(pvmsbufmid, pcbp->tt_tid, TC_EOC, &ztv);
	  if (cc < 0)
	  {
	      sprintf(pvmtxt, "ttpcb_flushall() TC_EOC mroute() to t%x failed, code %d\n", pcbp->tt_tid, cc);
	      pvmlogerror(pvmtxt);
	  }
	  pvm_freebuf(pvm_setsbuf(sbf));
      }
      else
      {
	  if (debugmask & TDMROUTE)
	  {
	      sprintf(pvmtxt, "ttpcb_flushall() closing not fully established t%x\n", pcbp->tt_tid);
	      pvmlogerror(pvmtxt);
	  }
	  pcbp = pcbp->tt_rlink;
	  ttpcb_delete(pcbp->tt_link);
      }
  }
  if (debugmask & TDMROUTE)
      pvmlogerror("ttpcb_flushall() direct connections list end\n");

  /* Loop until all the direct connections are properly closed.  Note that
     some TC_EOC messages may already be waiting, hence this "strange"
     code structure.

     Do the same for the connection to pvmd.  */
  while (ttlist->tt_link != ttlist || !received_eoc_from_pvmd)
  {
      fd_set rfds;
      char removed_something = 0;
      struct umbuf* up;
      struct timeval tmout = {5, 0};

      for (up = pvmrxlist->ub_link; up != pvmrxlist; up = up->ub_link)
      {
	  if (up->ub_cod == (int) TC_EOC)
	  {
	      if (debugmask & TDMROUTE)
	      {
		  sprintf(pvmtxt, "ttpcb_flushall() received TC_EOC from t%x\n", up->ub_src);
		  pvmlogerror(pvmtxt);
	      }
	      if (up->ub_src == TIDPVMD)
	      {
		  up = up->ub_rlink;
		  umbuf_free(up->ub_link->ub_mid);
		  received_eoc_from_pvmd = 1;
		  removed_something = 1;
		  continue;
	      }
	      pcbp = ttpcb_find(up->ub_src);
	      up = up->ub_rlink;
	      umbuf_free(up->ub_link->ub_mid);
	      if (pcbp)
	      {
		  if (debugmask & TDMROUTE)
		  {
		      sprintf(pvmtxt, "ttpcb_flushall() freeing pcb t%x\n", pcbp->tt_tid);
		      pvmlogerror(pvmtxt);
		  }
		  ttpcb_delete(pcbp);
		  removed_something = 1;
	      }
	  }
      }

      /* We don't want to call select() unless absolutely necessary.  */
      if (removed_something)
	  continue;

      rfds = pvmrfds;
      if (received_eoc_from_pvmd)
	  FD_CLR(topvmd->tt_fd, &rfds);

      if (debugmask & TDMSELECT)
	  print_fdset("ttpcb_flushall() rfds=", pvmnfds, &rfds);
      n = select(pvmnfds, &rfds, 0, 0, &tmout);
      if (debugmask & TDMSELECT)
      {
	  if (n < 0)
	      pvmlogperror("select");
	  sprintf(pvmtxt, "ttpcb_flushall() select returns %d\n", n);
	  pvmlogerror(pvmtxt);
      }
      if (n > 0)
      {
	  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
	      if (FD_ISSET(pcbp->tt_fd, &rfds))
	      {
		  if (mxinput(pcbp, 0) < 0)
		  {
		      if (debugmask & TDMROUTE)
		      {
			  sprintf(pvmtxt, "ttpcb_flushall() mxinput failed, freeing pcb t%x\n", pcbp->tt_tid);
			  pvmlogerror(pvmtxt);
		      }
		      pcbp = pcbp->tt_rlink;
		      ttpcb_delete(pcbp->tt_link);
		  }
	      }
	  if (FD_ISSET(topvmd->tt_fd, &rfds))
	  {
	      if (mxinput(topvmd, 0) < 0)
	      {
		  if (debugmask & TDMROUTE)
		      pvmlogerror("ttpcb_flushall() mxinput on topvmd failed\n");
		  received_eoc_from_pvmd = 1;
	      }
	  }
      }
      else
	  pvmlogerror("ttpcb_flushall() select TIMED OUT\n");
  }
#endif

  sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
  rbf = pvm_setrbuf(0);

  dummy = 0;
  pvm_pkint(&dummy, 1, 1);
#if 0
  msendrecv(TIDPVMD, TM_MIG);
#else
  ztv.tv_sec = ztv.tv_usec = 0;
  cc = mroute(pvmsbufmid, TIDPVMD, TM_MIG, &ztv);
  if (cc < 0)
  {
    sprintf(pvmtxt, "TM_MIG mroute() failed, code %d\n", cc);
    pvmlogerror(pvmtxt);
  }
#endif

  pvm_freebuf(pvm_setsbuf(sbf));
  pvm_setrbuf(rbf);

#if 0
  for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
    ttpcb_dead(pcbp);
#endif

 /* shutdown(topvmd->tt_fd, 1); */

  return 0;
}


/*
 * mroute()
 * 
 * Route a message to a destination. Returns when outgoing message (if
 * any) fully sent and (timed out (tmout) or at least one message
 * fully received) Returns >=0 the number of complete messages
 * downloaded, or negative on error.
 */

int
mroute(mid, dtid, code, tmout)
  int             mid;		/* message (or 0) */
  int             dtid;		/* dest */
  int             code;		/* type code */
  struct timeval *tmout;	/* get at least one message */
{
  struct ttpcb   *pcbp;		/* pcb for foreign task */
  int             s;		/* socket */
  int             sbf;		/* temp TC_CONREQ message */
  int             sbf2;		/* temp TM_NOTIFY message */
  struct umbuf   *up;
  struct sockaddr_in sad;
  int             l;
  int             cc;
  int             gotem = 0;
  static struct timeval ztv = {0, 0};
  char           *addr = 0;

#ifndef NOUNIXDOM
  struct sockaddr_un uns;
  char            spath[LEN_OF_TMP_NAM];

#endif

  if (debugmask & TDMROUTE)
  {
    if (tmout)
    {
      sprintf(pvmtxt, "mroute(%d, %x, %d, {%d, %d})\n",
	      mid, dtid, code, tmout->tv_sec, tmout->tv_usec);
    } else
    {
      sprintf(pvmtxt, "mroute(%d, %x, %d, NULL)\n",
	      mid, dtid, code);
    }
    pvmlogerror(pvmtxt);
  }
  if (up = midtobuf(mid))
  {
    up->ub_dst = dtid;
    up->ub_cod = code;

    if (pvmrouteopt == PvmRouteDirect
	&& dtid && TIDISTASK(dtid) && dtid != pvmmytid
	&& !ttpcb_find(dtid))
    {

#ifndef NOUNIXDOM
      /* XXX fvdl: route lookup */

      if ((pvmmytid & pvmtidhmask) == (dtid & pvmtidhmask))
      {
	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
	{
	  pvmlogperror("mroute() socket");

	} else
	{
	  (void) TMPNAMFUN(spath);
	  BZERO((char *) &uns, sizeof(uns));
	  uns.sun_family = AF_UNIX;
	  spath[0] = 0;
	  (void) tmpnam(spath);
	  strcpy(uns.sun_path, spath);

	  if (bind(s, (struct sockaddr *) & uns, sizeof(uns)) == -1)
	  {
	    pvmlogperror("mroute() bind");
	    (void) close(s);
	    s = -1;

	  } else
	  {
	    pcbp = ttpcb_creat(dtid);
	    pcbp->tt_fd = s;
	    addr = pcbp->tt_spath = STRALLOC(spath);
	    pcbp->tt_state = TTCONWAIT;
	  }
	}

      } else
      {
#endif				/* NOUNIXDOM */

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
	  pvmlogperror("mroute() socket");

	} else
	{
	  sad = pvmourinet;
	  l = sizeof(sad);
	  if (bind(s, (struct sockaddr *) & sad, sizeof(sad)) == -1)
	  {
	    pvmlogperror("mroute() bind");
	    (void) close(s);
	    s = -1;

	  } else
	  {
	    if (getsockname(s, (struct sockaddr *) & sad, &l) == -1)
	    {
	      pvmlogperror("mroute() getsockname");
	      (void) close(s);
	      s = -1;

	    } else
	    {
#ifndef NOSOCKOPT
	      l = 1;
	      if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
			     (char *) &l, sizeof(int)) == -1)
		pvmlogperror("mroute() setsockopt");
#ifndef NOSOCKBUF
	      l = TTSOCKBUF;
	      if (setsockopt(s, SOL_SOCKET, SO_SNDBUF,
			     (char *) &l, sizeof(l)) == -1
		  || setsockopt(s, SOL_SOCKET, SO_RCVBUF,
				(char *) &l, sizeof(l)) == -1)
		pvmlogperror("mroute() setsockopt SO_SNDBUF");
#endif
#endif
	      pcbp = ttpcb_creat(dtid);
	      pcbp->tt_fd = s;
	      pcbp->tt_sad = sad;
	      pcbp->tt_state = TTCONWAIT;
	      addr = inadport_hex(&sad);
	    }
	  }
	}
#ifndef NOUNIXDOM
      }
#endif

      if (s >= 0)
      {
	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	l = TDPROTOCOL;
	pvm_pkint(&l, 1, 1);
	pvm_pkstr(addr);
	sbf = pvm_setsbuf(sbf);
	up = midtobuf(sbf);
	up->ub_dst = dtid;
	up->ub_cod = TC_CONREQ;

	sbf2 = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	l = PvmTaskExit;
	pvm_pkint(&l, 1, 1);
	l = TC_TASKEXIT;
	pvm_pkint(&l, 1, 1);
	l = 1;
	pvm_pkint(&l, 1, 1);
	pvm_pkint(&dtid, 1, 1);
	sbf2 = pvm_setsbuf(sbf2);
	up = midtobuf(sbf2);
	up->ub_dst = TIDPVMD;
	up->ub_cod = TM_NOTIFY;

	if (debugmask & TDMROUTE)
	{
	  sprintf(pvmtxt, "mroute() CONREQ to t%x\n", dtid);
	  pvmlogerror(pvmtxt);
	}
	/*
	 * XXX it would be better if mxfer() took a list of send
	 * messages
	 */
	do
	{
	  if ((cc = mxfer(sbf,
			  (sbf2 ? &ztv : (struct timeval *) 0))) < 0)
	    break;		/* XXX hmm */
	  gotem += cc;
	  if (sbf)
	  {
	    pvm_freebuf(sbf);
	    if (sbf2)
	    {
	      sbf = sbf2;
	      sbf2 = 0;
	    } else
	      sbf = 0;
	  }
	} while (
#ifdef DPVM
		 /* If the task is migrated in mxfer() above, then pcbp pointer
		    is no longer valid.  */
		 (pcbp = ttpcb_find(dtid)) &&
#endif
		 pcbp->tt_state == TTCONWAIT);
      }
    }
  }
  cc = mxfer(mid, tmout);
  pcbp = ttlist->tt_link;
  while (pcbp != ttlist)
    if (pcbp->tt_state == TTDEAD)
    {
      if (debugmask & TDMROUTE)
      {
	sprintf(pvmtxt, "mroute() freeing pcb t%x\n", pcbp->tt_tid);
	pvmlogerror(pvmtxt);
      }
      pcbp = pcbp->tt_link;
      ttpcb_delete(pcbp->tt_rlink);
    } else
      pcbp = pcbp->tt_link;
  return (cc < 0 ? cc : gotem + cc);
}


/*
 * pvm_tc_conreq()
 * 
 * Another task requests a connection with us.  Reply with CONACK
 * message.
 * 
 * Returns mid of reply.
 * 
 * TC_CONREQ() { int tdprotocol		// t-d protocol revision
 * number string sockaddr		// address of other socket }
 */

static int
pvm_tc_conreq(up)
  struct umbuf   *up;
{
  int             sbf = 0;	/* reply message mid */
  int             ttpro;	/* protocol revision */
  struct ttpcb   *pcbp;		/* pcb for connection */
  int             ackd;		/* allow connection (0) */
  char           *addr;		/* socket address */
  static int      linger[2] = {1, 60};	/* XXX arbitrary time */
  int             i;
  struct umbuf   *up2;

#ifndef NOUNIXDOM
  struct sockaddr_un uns;
  char            spath[LEN_OF_TMP_NAM];

#endif
  char            buf[256];

  pvm_upkint(&ttpro, 1, 1);
  pvm_upkstr(buf);

  if (pcbp = ttpcb_find(up->ub_src))
  {
    if (pcbp->tt_state == TTDEAD)
    {
      ttpcb_delete(pcbp);
      goto recreate;
    }
    if (debugmask & TDMROUTE)
    {
      sprintf(pvmtxt, "pvmmctl() crossed CONREQ from t%x\n",
	      up->ub_src);
      pvmlogerror(pvmtxt);
    }
    if (pcbp->tt_state == TTCONWAIT)
    {
#if 0 /* This code is hopelessly broken, instead use a back-ported code from
	 PVM 3.4.  */
      if (buf[0] == '/')
      {
#ifdef NOUNIXDOM
	sprintf(pvmtxt,
		"pvmmctl() CONREQ from t%x, Unix domain socket unsupported\n",
		up->ub_src);
	pvmlogerror(pvmtxt);

#else				/* NOUNIXDOM */
	BZERO((char *) &uns, sizeof(uns));
	uns.sun_family = AF_UNIX;
	strcpy(uns.sun_path, buf);
	while ((i = connect(pcbp->tt_fd, (struct sockaddr *) & uns,
			    sizeof(uns))) == -1
	       && errno == EINTR)
	  ;
	if (i == -1)
	  pvmlogperror("pvmmctl() connect");
	else
	  pcbp->tt_state = TTOPEN;
#endif				/* NOUNIXDOM */

      } else
      {
	pcbp->tt_osad.sin_family = AF_INET;
	hex_inadport(buf, &pcbp->tt_osad);
	while ((i = connect(pcbp->tt_fd,
			    (struct sockaddr *) & pcbp->tt_osad,
			    sizeof(pcbp->tt_osad))) == -1
	       && errno == EINTR)
	  ;
	if (i == -1)
	{
	  pvmlogperror("pvmmctl() connect");

	} else
	{
	  pcbp->tt_state = TTOPEN;
#ifndef NOSOCKOPT
	  if (setsockopt(pcbp->tt_fd, SOL_SOCKET, SO_LINGER,
			 (char *) linger, sizeof(linger)) == -1)
	    pvmlogperror("pvmmctl() setsockopt");
#endif				/* NOSOCKOPT */
	}

      }
      if (pcbp->tt_state == TTOPEN)
      {
#ifndef DPVM
	  /* Non-blocking communication is deadly for DPVM.  For example,
	     its TC_EOC messages could be interpreted as part of data if
	     write() call of data was split.  */
	if ((i = fcntl(pcbp->tt_fd, F_GETFL, 0)) == -1)
	  pvmlogperror("pvmmctl() fcntl");
	else
	{
#ifdef O_NDELAY
	  i |= O_NDELAY;
#else
	  i |= FNDELAY;
#endif
	  (void) fcntl(pcbp->tt_fd, F_SETFL, i);
	}
#endif /* ! DPVM */
	pvm_fd_add(pcbp->tt_fd, 1);
	dpvm_set(pcbp->tt_fd);
      }
#else

			/* must handle simultaneous connect from both ends:
			 *  lower tid of the two just ignores the CONREQ.
			 *  higher tid pretends it didn't send a CONREQ,
			 *      sends a CONACK back.
			 * this makes the connection single-sided.
			 */

			if (debugmask & TDMROUTE)
				pvmlogerror( "pvmmctl() handling crossed CONREQ\n");

			if (pvmmytid > up->ub_src) {
				if (listen(pcbp->tt_fd, 1) == -1)
					pvmlogperror("pvm_tc_conreq() listen");

				else {
					pcbp->tt_state = TTGRNWAIT;
					pvm_fd_add(pcbp->tt_fd, 1);
					ackd = 0;
					if (buf[0] == '/') {
#ifdef NOUNIXDOM
	sprintf(pvmtxt,
		"pvm_tc_conreq() CONREQ from t%x, Unix domain socket unsupported\n",
		up->ub_src);
	pvmlogerror(pvmtxt);
#else
						addr = pcbp->tt_spath;
#endif
					} else {
						hex_inadport(buf, &pcbp->tt_osad);
						addr = inadport_hex(&pcbp->tt_sad);
					}

					sbf=pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
					ttpro = TDPROTOCOL;
					pvm_pkint(&ttpro, 1, 1);
					pvm_pkint(&ackd, 1, 1);
					pvm_pkstr(addr);
    sbf = pvm_setsbuf(sbf);
    up2 = midtobuf(sbf);
    up2->ub_dst = up->ub_src;
    up2->ub_cod = TC_CONACK;
				}
			}

#endif
#ifdef DPVM
    } else if (pcbp->tt_state == TTGRNWAIT)
    {
	if (debugmask & TDMROUTE)
	    pvmlogerror("pvmmctl() handling second CONREQ in TTGRNWAIT state\n");
	hex_inadport(buf, &pcbp->tt_osad);
	ackd = 0;
	addr = inadport_hex(&pcbp->tt_sad);

	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	ttpro = TDPROTOCOL;
	pvm_pkint(&ttpro, 1, 1);
	pvm_pkint(&ackd, 1, 1);
	pvm_pkstr(addr);
	sbf = pvm_setsbuf(sbf);
	up2 = midtobuf(sbf);
	up2->ub_dst = up->ub_src;
	up2->ub_cod = TC_CONACK;
#endif
    } else
    {
      sprintf(pvmtxt, "pvmmctl() CONREQ from t%x but state=%d ?\n",
	      up->ub_src, pcbp->tt_state);
      pvmlogerror(pvmtxt);
    }

  } else
  {
recreate:
    if (debugmask & TDMROUTE)
    {
      sprintf(pvmtxt, "pvmmctl() CONREQ from t%x\n", up->ub_src);
      pvmlogerror(pvmtxt);
    }
    ackd = 1;
    pcbp = ttpcb_creat(up->ub_src);
    if (pvmrouteopt != PvmDontRoute)
    {
      if (buf[0] == '/')
      {
#ifdef NOUNIXDOM
	sprintf(pvmtxt,
		"pvmmctl() CONREQ from t%x, Unix domain socket unsupported\n",
		up->ub_src);
	pvmlogerror(pvmtxt);

#else				/* NOUNIXDOM */
	if ((pcbp->tt_fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
	{
	  pvmlogperror("pvmmctl() socket");

	} else
	{
	  (void) TMPNAMFUN(spath);
	  BZERO((char *) &uns, sizeof(uns));
	  uns.sun_family = AF_UNIX;
	  spath[0] = 0;
	  (void) tmpnam(spath);
	  strcpy(uns.sun_path, spath);

	  if (bind(pcbp->tt_fd, (struct sockaddr *) & uns, sizeof(uns))
	      == -1)
	  {
	    pvmlogperror("pvmmctl() bind");

	  } else
	  {
	    if (listen(pcbp->tt_fd, 1) == -1)
	      pvmlogperror("pvmmctl() listen");

	    else
	    {
	      pcbp->tt_state = TTGRNWAIT;
	      pvm_fd_add(pcbp->tt_fd, 1);
	      ackd = 0;
	      addr = pcbp->tt_spath = STRALLOC(spath);
	    }
	  }
	}
#endif				/* NOUNIXDOM */

      } else
      {
	if ((pcbp->tt_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
	  pvmlogperror("pvmmctl() socket");

	} else
	{
	  pcbp->tt_sad = pvmourinet;
	  i = sizeof(pcbp->tt_sad);
	  if (bind(pcbp->tt_fd, (struct sockaddr *) & pcbp->tt_sad, i)
	      == -1)
	  {
	    pvmlogperror("pvmmctl() bind");

	  } else
	  {
	    if (getsockname(pcbp->tt_fd,
		      (struct sockaddr *) & pcbp->tt_sad, &i) == -1)
	    {
	      pvmlogperror("pvmmctl() getsockname");

	    } else
	    {
	      if (listen(pcbp->tt_fd, 1) == -1)
		pvmlogperror("pvmmctl() listen");

	      else
	      {
		hex_inadport(buf, &pcbp->tt_osad);
		pcbp->tt_state = TTGRNWAIT;
		pvm_fd_add(pcbp->tt_fd, 1);
		ackd = 0;
		addr = inadport_hex(&pcbp->tt_sad);
	      }
	    }
	  }
	}
      }
    }
    sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
    ttpro = TDPROTOCOL;
    pvm_pkint(&ttpro, 1, 1);
    pvm_pkint(&ackd, 1, 1);
    pvm_pkstr(addr);
    sbf = pvm_setsbuf(sbf);
    up2 = midtobuf(sbf);
    up2->ub_dst = up->ub_src;
    up2->ub_cod = TC_CONACK;

    if (ackd)
      ttpcb_delete(pcbp);
  }
  return sbf;
}


/*
 * pvm_tc_conack()
 * 
 * Another task replies to our connection request.
 * 
 * TC_CONACK() { int tdprotocol		// t-d protocol revision
 * number int ack				// 0 ok, 1 denied
 * string sockaddr		// address of other socket }
 */

static int
pvm_tc_conack(up)
  struct umbuf   *up;
{
  int             ttpro;	/* protocol revision */
  int             ackd;		/* connection allowed (0) */
  struct ttpcb   *pcbp;		/* pcb for connection */
  static int      linger[2] = {1, 60};	/* XXX arbitrary time */
  int             i;
#ifdef DPVM
  int             connect_failed = 0; /* marks that it is connect that failed */
#endif

#ifndef NOUNIXDOM
  struct sockaddr_un uns;

#endif
  char            buf[256];

  pvm_upkint(&ttpro, 1, 1);
  pvm_upkint(&ackd, 1, 1);
  pvm_upkstr(buf);

  if (pcbp = ttpcb_find(up->ub_src))
  {
    if (pcbp->tt_state == TTCONWAIT)
    {
      if (debugmask & TDMROUTE)
      {
	sprintf(pvmtxt, "pvmmctl() CONACK from t%x\n", up->ub_src);
	pvmlogerror(pvmtxt);
      }
      if (ttpro != TDPROTOCOL)
      {
	sprintf(pvmtxt,
		"pvmmctl() t-t protocol mismatch with t%x\n",
		pcbp->tt_tid);
	pvmlogerror(pvmtxt);
	ackd = 1;

      } else
      {
	if (ackd != 0)
	{
	  if (debugmask & TDMROUTE)
	  {
	    sprintf(pvmtxt, "pvmmctl() route to t%x denied\n",
		    pcbp->tt_tid);
	    pvmlogerror(pvmtxt);
	  }
	} else
	{
	  if (buf[0] == '/')
	  {
#ifdef NOUNIXDOM
	    sprintf(pvmtxt,
		    "pvmmctl() CONREQ from t%x, Unix domain socket unsupported\n",
		    up->ub_src);
	    pvmlogerror(pvmtxt);
	    ackd = 1;
#else				/* NOUNIXDOM */
	    BZERO((char *) &uns, sizeof(uns));
	    uns.sun_family = AF_UNIX;
	    strcpy(uns.sun_path, buf);
	    while ((i = connect(pcbp->tt_fd, (struct sockaddr *) & uns,
				sizeof(uns))) == -1
		   && errno == EINTR)
	      ;
	    if (i == -1)
	    {
	      pvmlogperror("pvmmctl() connect");
	      ackd = 1;

	    } else
	    {
	      pcbp->tt_state = TTOPEN;
	      if ((i = fcntl(pcbp->tt_fd, F_GETFL, 0)) == -1)
		pvmlogperror("pvmmctl() fcntl");
	      else
	      {
#ifdef O_NDELAY
		i |= O_NDELAY;
#else
		i |= FNDELAY;
#endif
		(void) fcntl(pcbp->tt_fd, F_SETFL, i);
	      }
	      pvm_fd_add(pcbp->tt_fd, 1);
	      dpvm_set(pcbp->tt_fd);
	    }

#endif				/* NOUNIXDOM */

	  } else
	  {
	    pcbp->tt_osad.sin_family = AF_INET;
	    hex_inadport(buf, &pcbp->tt_osad);
	    while ((i = connect(pcbp->tt_fd,
				(struct sockaddr *) & pcbp->tt_osad,
				sizeof(pcbp->tt_osad))) == -1
		   && errno == EINTR)
	      ;
	    if (i == -1)
	    {
	      pvmlogperror("pvmmctl() connect");
	      ackd = 1;
#ifdef DPVM
	      connect_failed = 1;
#endif
	    } else
	    {
	      pcbp->tt_state = TTOPEN;
#ifndef NOSOCKOPT
	      if (setsockopt(pcbp->tt_fd, SOL_SOCKET, SO_LINGER,
			     (char *) linger, sizeof(linger)) == -1)
		pvmlogperror("pvmmctl() setsockopt");
#endif				/* NOSOCKOPT */
#ifndef DPVM
	      /* Non-blocking communication is deadly for DPVM.  For example,
		 its TC_EOC messages could be interpreted as part of data if
		 write() call of data was split.  */
	      if ((i = fcntl(pcbp->tt_fd, F_GETFL, 0)) == -1)
		pvmlogperror("pvmmctl() fcntl");
	      else
	      {
#ifdef O_NDELAY
		i |= O_NDELAY;
#else
		i |= FNDELAY;
#endif
		(void) fcntl(pcbp->tt_fd, F_SETFL, i);
	      }
#else /* DPVM */
	      /* Read the start-of-connection message.  */
	      for (;;)
	      {
		  fd_set fds;
		  struct timeval tmout = {5, 0};
		  int n, maxfd;
		  struct ttpcb* tt;

		  FD_ZERO(&fds);
		  FD_SET(pcbp->tt_fd, &fds);
		  maxfd = pcbp->tt_fd;
		  /* Add all the file descriptors of direct connections that
		     are being established.  We must handle them here, or
		     deadlocks can result if there is a cycle of processes
		     wishing to communicate.  */
		  for (tt = ttlist->tt_link; tt != ttlist; tt = tt->tt_link)
		      if (tt->tt_state == TTGRNWAIT)
		      {
			  FD_SET(tt->tt_fd, &fds);
			  if (maxfd < tt->tt_fd)
			      maxfd = tt->tt_fd;
		      }

		  n = select(maxfd + 1, &fds, 0, 0, &tmout);
		  if (debugmask & TDMSELECT)
		  {
		      if (n < 0)
			  pvmlogperror("select");
		      sprintf(pvmtxt, "pvm_tc_conack() select returns %d\n", n);
		      pvmlogerror(pvmtxt);
		  }
		  if (n > 0)
		  {
		      if (FD_ISSET(pcbp->tt_fd, &fds))
		      {
			  unsigned char c;
			  n = read(pcbp->tt_fd, &c, 1);
			  if (debugmask & TDMROUTE)
			  {
			      if (n < 0)
				  pvmlogperror("read");
			      sprintf(pvmtxt,"pvm_tc_conack() read returned %d\n", n);
			      pvmlogerror(pvmtxt);
			  }
			  if (n == 1)
			  {
			      if (c == TT_SOC)
			      {
				  if (debugmask & TDMROUTE)
				      pvmlogerror("pvm_tc_conack() TT_SOC received\n");
			      }
			      else
				  pvmlogerror("pvm_tc_conack() first byte not TT_SOC\n");
			      break;
			  }
			  else
			      break; /* Connection closed by the other side.  */
		      }

		      for (tt = ttlist->tt_link; tt != ttlist; tt = tt->tt_link)
			  if (tt->tt_state == TTGRNWAIT &&
			      FD_ISSET(tt->tt_fd, &fds))
			  {
			      /* !!! WARNING! Code duplication.

				 The code below is shamelessly copied from
				 mxfer, and slightly adjusted to fit here.  */
			      int n;
			      int s;
			      struct sockaddr_in sad;
			      struct ttpcb* pcbp = tt;

	n = sizeof(sad);
	s = accept(pcbp->tt_fd, (struct sockaddr *) & sad, &n);
	if (s == -1)
	{
	  pvmlogperror("pvm_tc_conack() accept");
	  ttpcb_dead(pcbp);

	} else
	{
#ifndef NOSOCKOPT
#ifndef NOUNIXDOM
	  if (!pcbp->tt_spath)
	  {			/* not unix sock */
#endif
	    static int linger[2] = {1, 60};
	    n = 1;
	    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
			   (char *) &n, sizeof(int)) == -1)
	      pvmlogperror("pvm_tc_conack() setsockopt");
	    if (setsockopt(s, SOL_SOCKET, SO_LINGER,
			   (char *) linger, sizeof(linger)) == -1)
	      pvmlogperror("pvm_tc_conack() setsockopt");
#ifndef NOSOCKBUF
	    n = TTSOCKBUF;
	    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF,
			   (char *) &n, sizeof(int)) == -1
		|| setsockopt(s, SOL_SOCKET, SO_RCVBUF,
			      (char *) &n, sizeof(n)) == -1)
	      pvmlogperror("pvm_tc_conack() setsockopt SO_SNDBUF");
#endif
#ifndef NOUNIXDOM
	  }
#endif
#endif				/* NOSOCKOPT */
	  if (debugmask & TDMROUTE)
	  {
	    sprintf(pvmtxt, "pvm_tc_conack() accept from t%x\n",
		    pcbp->tt_tid);
	    pvmlogerror(pvmtxt);
	  }
	  (void) close(pcbp->tt_fd);
	  /*
	   * (void)dup2(s, pcbp->tt_fd); (void)close(s);
	   */
	  pvm_fd_delete(pcbp->tt_fd, 1);
	  pcbp->tt_fd = s;
	  pcbp->tt_state = TTOPEN;
#ifdef DPVM
	  {
	      /* Send the start-of-connection message.  To avoid any further
		 races and tricky cases, we send a crude message not adhearing
		 to PVM standards, but at least we send it immediately, without
		 a danger of a signal handler being invoked in between.  */
	      char c = TT_SOC;
	      if (write(pcbp->tt_fd, &c, 1) != 1)
		  ttpcb_dead(pcbp);
	  }
#endif
	  pvm_fd_add(s, 1);
	  dpvm_set(pcbp->tt_fd);
	}
			  }
		  }
	      }
#endif /* DPVM */
	      pvm_fd_add(pcbp->tt_fd, 1);
	      dpvm_set(pcbp->tt_fd);
	    }
	  }
	}
      }

      if (ackd != 0)
#ifdef DPVM
      if (connect_failed)
	  /* This is very likely if a migration happened in the mean time. So
	     instead of marking the connection as semi-permanently denied, we
	     just abort for now, but will retry on the next occasion.  */
	  ttpcb_dead(pcbp);
      else
#endif
      {
	pcbp->tt_state = TTDENY;
	(void) close(pcbp->tt_fd);
	pcbp->tt_fd = -1;
      }
    } else
    {
      sprintf(pvmtxt, "pvmmctl() CONACK from t%x but state=%d\n",
	      up->ub_src, pcbp->tt_state);
      pvmlogerror(pvmtxt);
    }

  } else
  {
    sprintf(pvmtxt, "pvmmctl() suprious CONACK from t%x\n",
	    up->ub_src);
    pvmlogerror(pvmtxt);
  }
  return 0;
}


/*
 * pvmmctl()
 * 
 * Entry points for libpvm control messages.
 * 
 * Called by mxinput() when a control message (code between TC_FIRST and
 * TC_LAST) is received.
 */

static int
pvmmctl(up)
  struct umbuf   *up;
{
  int             rbf;		/* temp rx message storage */
  int             sbf = 0;	/* return message to send */
  struct ttpcb   *pcbp;
  int             tid;
  char            buf[256];

  rbf = pvm_setrbuf(up->ub_mid);

  switch (up->ub_cod)
  {

      /*
       * do the messy ones in separate functions
       */

    case TC_CONREQ:
      sbf = pvm_tc_conreq(up);
      break;

    case TC_CONACK:
      sbf = pvm_tc_conack(up);
      break;

      /*
       * we are notified that another task (to which we have a direct
       * route) has exited
       */

#ifdef DPVM
    case TC_MOVED: /* This is only possible if somebody disabled the
		      SIGURG signal handler. We do it in the PVM console
		      and DPVM monitors. */
#endif
    case TC_TASKEXIT:
      pvm_upkint(&tid, 1, 1);

      if (debugmask & TDMROUTE)
      {
	sprintf(pvmtxt, "pvmmctl() TASKEXIT or MOVED for t%x\n", tid);
	pvmlogerror(pvmtxt);
      }

      /*
       * v3.3: for now, don't close fully open route as we may lose
       * messages.  rely on socket getting EOF
       */
      if (pcbp = ttpcb_find(tid))
      {
	  if (pcbp->tt_state != TTOPEN)
	      ttpcb_dead(pcbp);
#if 0 /* This code should not be needed, dpvm_oobhandler should always be
	 invoked first.  This was supposed to be a workaround for a bug,
	 but the source of the bug was found to be different.  */
	  else if (up->ub_cod == TC_MOVED)
	  {
	      struct umbuf *up2;
	      int dummy;

	      if (debugmask & TDMROUTE)
	      {
		  sprintf(pvmtxt, "pvmmctl(): t%x about to migrate\n", tid);
		  pvmlogerror(pvmtxt);
	      }

	      /* We can't allow mxfer() to send any more data via this socket.  */
	      pcbp->tt_migrating = 1;

	      /* Prepare end of connection marker message.  */
	      sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	      dummy = 0;
	      pvm_pkint(&dummy, 1, 1);
	      sbf = pvm_setsbuf(sbf);
	      up2 = midtobuf(sbf);
	      up2->ub_dst = tid;
	      up2->ub_cod = TC_EOC;

	      /* Yet another awful hack to fix dpvm_oobhandler() races.
		 Theoretically, dpvm_oobhandler() should be invoked before
		 the TC_MOVED message arrives, as the signal is sent before
		 sending the message. Apparently, this is not always the case,
		 particularly with "heavy" processes under Solaris. So we
		 notify the signal handler that it needs not do anything. */
	      tc_moved_handled_by_pvmmctl++;
	  }
#endif
      }
      break;

      /*
       * we get output from a child* task (because we called
       * pvm_catchout())
       * 
       * TC_OUTPUT() { int tid				// of source
       * task int length			// of output char
       * output[length]	// characters printed }
       */

    case TC_OUTPUT:
      pvmclaimo();
      break;

      /*
       * just like the name
       */

    case TC_NOOP:
      if (debugmask)
      {
	sprintf(pvmtxt, "pvmmctl() NOOP from t%x\n", up->ub_src);
	pvmlogerror(pvmtxt);
      }
      break;

      /*
       * set our trace mask on the fly
       * 
       * TC_SETTMASK() { string mask			// new trace
       * mask }
       */

    case TC_SETTMASK:
      pvm_upkstr(buf);
      if (strlen(buf) + 1 == TEV_MASK_LENGTH)
	BCOPY(buf, pvmtrcmask, TEV_MASK_LENGTH);
      else
	pvmlogerror("pvmmctl() bogus trace mask\n");
      break;

    default:
      sprintf(pvmtxt, "pvmmctl() from t%x code=%d ?\n",
	      up->ub_src, up->ub_cod);
      pvmlogerror(pvmtxt);
      break;
  }

  pvm_setrbuf(rbf);
  umbuf_free(up->ub_mid);
  return sbf;
}


/*
 * mxinput()
 * 
 * Input from a connection. Reassemble packets and messages and add them
 * to pvmrxlist. Returns >= 0 the number of messages added to
 * pvmrxlist, or negative on error.
 */

int
mxinput(pcbp, dontblock)
  struct ttpcb   *pcbp;
  int             dontblock;

{
  int             gotem = 0;	/* num msgs added to pvmrxlist */
  struct frag    *fp;		/* shadows pcbp->tt_rxf */
  struct frag    *fp2;
  int             n;		/* bytes received */
  int             m;		/* length of fragment */
  struct umbuf   *rxup;		/* message containing this frag */
  char           *cp;		/* gp */
  int             src;
  int             len;
  int             ff;

#ifdef DPVM
  int             msgid, seq;

#endif

  if (!pcbp->tt_rxf)
  {
    if (debugmask & TDMPACKET)
      pvmlogerror("mxinput() new fragment\n");
    pcbp->tt_rxf = fr_new(pvmfrgsiz);
  }
  fp = pcbp->tt_rxf;

  /* read fragment header separately from body */

  n = (fp->fr_len < TDFRAGHDR) ? 0 : pvmget32(fp->fr_dat + 8);
  n += TDFRAGHDR - fp->fr_len;
  if (debugmask & TDMPACKET)
  {
    sprintf(pvmtxt, "mxinput() pcb t%x fr_len=%d fr_dat=+%d n=%d\n",
	    pcbp->tt_tid, fp->fr_len, fp->fr_dat - fp->fr_buf, n);
    pvmlogerror(pvmtxt);
  }
  n = read(pcbp->tt_fd, fp->fr_dat + fp->fr_len, n);
  if (debugmask & TDMPACKET)
  {
    sprintf(pvmtxt, "mxinput() read=%d\n", n);
    pvmlogerror(pvmtxt);
  }
  if (n == -1 && errno == EWOULDBLOCK && dontblock)
  {
    if (debugmask & TDMPACKET)
      pvmlogerror("mxinput(): EWOULDBLOCK && dontblock=1\n");
    return -1;
  }
  /* deal with errors and closes */

  if (n == -1 && errno != EWOULDBLOCK && errno != EINTR)
  {
    if (debugmask & TDMPACKET)
    {
      sprintf(pvmtxt, "mxinput() t%x read", pcbp->tt_tid);
      pvmlogperror(pvmtxt);
    }
    return PvmSysErr;
  }
  if (!n)
  {
    if (debugmask & TDMPACKET)
    {
      sprintf(pvmtxt, "mxinput() t%x read EOF\n", pcbp->tt_tid);
      pvmlogerror(pvmtxt);
    }
    return -1;
  }
  if (n < 1)
    return gotem;
  if ((fp->fr_len += n) < TDFRAGHDR)
    return gotem;

  /* realloc buffer if frag won't fit */

  m = TDFRAGHDR + pvmget32(fp->fr_dat + 8);
  if (fp->fr_len == TDFRAGHDR)
  {
    if (m > fp->fr_max - (fp->fr_dat - fp->fr_buf))
    {
      fp2 = fr_new(m);
      BCOPY(fp->fr_dat, fp2->fr_dat, TDFRAGHDR);
      fp2->fr_len = fp->fr_len;
      fr_unref(fp);
      fp = pcbp->tt_rxf = fp2;
      if (debugmask & TDMPACKET)
      {
	sprintf(pvmtxt, "mxinput() realloc frag max=%d\n", m);
	pvmlogerror(pvmtxt);
      }
    }
  }
  if (fp->fr_len == m)
  {
    pcbp->tt_rxf = 0;

    cp = fp->fr_dat;
    /*
     * dst = pvmget32(cp);		XXX might unpack and check
     */
    src = pvmget32(cp + 4);
    len = pvmget32(cp + 8);
#ifdef DPVM
    ff = pvmget8(cp + 16);
    fp->fr_len -= TDFRAGHDR;
    fp->fr_dat += TDFRAGHDR;
    cp += TDFRAGHDR;

    fp->fr_len -= TTFRAGHDR;
    fp->fr_dat += TTFRAGHDR;
    msgid = pvmget32(cp);
    fp->fr_seq = seq = pvmget32(cp + 4);
#else
    ff = pvmget8(cp + 12);
    fp->fr_len -= TDFRAGHDR;
    fp->fr_dat += TDFRAGHDR;
    cp += TDFRAGHDR;
#endif

    if (debugmask & TDMPACKET)
    {
      sprintf(pvmtxt, "mxinput() pkt src t%x len %d ff %d msgid %d seq %d\n",
	      src, fp->fr_len, ff, msgid, seq);
      pvmlogerror(pvmtxt);
    }
#ifdef DPVM
    for (rxup = pcbp->tt_rxfrag->ub_link; rxup != pcbp->tt_rxfrag;
	 rxup = rxup->ub_link)
      if (rxup->ub_src == src && rxup->ub_fid == msgid)
	break;
    if (rxup == pcbp->tt_rxfrag)
    {
      if (debugmask & TDMPACKET)
      {
	sprintf(pvmtxt, "mxinput() new msg id %d src %x\n",
		msgid, src);
	pvmlogerror(pvmtxt);
      }
      /*
       * No message yet, allocate a new one.
       */
      rxup = midtobuf(umbuf_new());
      rxup->ub_cod = rxup->ub_enc = rxup->ub_wid = -1;
      rxup->ub_src = src;
      rxup->ub_fid = msgid;
      rxup->ub_tlen = -1;
      LISTPUTBEFORE(pcbp->tt_rxfrag, rxup, ub_link, ub_rlink);
    }
    if (ff & FFSOM)
    {
      cp += TTFRAGHDR;
      fp->fr_len -= TTMSGHDR;
      fp->fr_dat += TTMSGHDR;
      rxup->ub_cod = pvmget32(cp);
      rxup->ub_enc = pvmget32(cp + 4);
      rxup->ub_wid = pvmget32(cp + 8);
      rxup->ub_tlen = pvmget32(cp + 12);
    }
    for (fp2 = rxup->ub_frag->fr_link; fp2 != rxup->ub_frag;
	 fp2 = fp2->fr_link)
	if (fp2->fr_seq >= fp->fr_seq)
	    break;
    if (fp->fr_seq == fp2->fr_seq)
    {
      sprintf(pvmtxt, "mxfer() dropped dup pkt src %x id %d seq %d\n",
	      src, msgid, fp->fr_seq);
      pvmlogerror(pvmtxt);
      fr_unref(fp);
    } else
    {
      LISTPUTBEFORE(fp2, fp, fr_link, fr_rlink);
      rxup->ub_len += fp->fr_len;
      /*
       * if we got all of it, move to pvmrxlist and count it
       */
      if (debugmask & PDMPACKET)
      {
	sprintf(pvmtxt, "mxinput() frlen %d tlen %d len %d wid %x\n",
	     fp->fr_len, rxup->ub_tlen, rxup->ub_len, rxup->ub_wid);
	pvmlogerror(pvmtxt);
      }
      if (rxup->ub_len == rxup->ub_tlen)
      {
	  /* Messages from PVM daemons are not sequenced.  Neither are TC_EOC
	     messages.  */
	  if (!(rxup->ub_src & TIDPVMD) && rxup->ub_cod != TC_EOC &&
	      rxup->ub_fid != tseq_get_lastrecv(rxup->ub_src) + 1)
	  {
	      if (debugmask & PDMPACKET)
	      {
		  sprintf(pvmtxt, "mxinput() out-of-seq msg, id %d, last %d\n",
			  rxup->ub_fid, tseq_get_lastrecv(rxup->ub_src));
		  pvmlogerror(pvmtxt);
	      }

	      /* The message is out of sequence, so we cannot move it into
		 pvmrxlist just yet.  We can't leave it in pcbp queue either,
		 since otherwise messages that arrive throught different pcbps
		 won't see each other, and besides direct pcbps are freed on
		 migration, so the message would get lost.  */

	      LISTDELETE(rxup, ub_link, ub_rlink);
	      LISTPUTBEFORE(oosrxlist, rxup, ub_link, ub_rlink);
	  }
	  else
	  do
	  {
	    struct umbuf* rxup2;

	    LISTDELETE(rxup, ub_link, ub_rlink);
	    rxup->ub_codef = enctovec(rxup->ub_enc);
	    LISTPUTBEFORE(pvmrxlist, rxup, ub_link, ub_rlink);
	    gotem++;

	    if (debugmask & PDMPACKET)
	    {
		sprintf(pvmtxt,
			"mxinput() src t%x route t%x cod %d len %d\n",
			rxup->ub_src, pcbp->tt_tid,
			rxup->ub_cod, rxup->ub_len);
		pvmlogerror(pvmtxt);
	    }
	    if ((rxup->ub_src & TIDPVMD) || rxup->ub_cod == TC_EOC)
		break;

	    tseq_increase_lastrecv(rxup->ub_src);

	    /* Scan for previously received out-of-sequence messages.  If
	       they are in sequence now, put them into pvmrxlist.  */
	    for (rxup2 = oosrxlist->ub_link; rxup2 != oosrxlist;
		 rxup2 = rxup2->ub_link)
		if (rxup2->ub_src == src && rxup2->ub_fid == rxup->ub_fid + 1)
		    break;
	    if ((debugmask & PDMPACKET) &&
		rxup2->ub_src == src && rxup2->ub_fid == rxup->ub_fid + 1)
	    {
		sprintf(pvmtxt, "mxinput() putting back in sequence msg id %d\n",
			rxup2->ub_fid);
		pvmlogerror(pvmtxt);
	    }
	    rxup = rxup2;
	  } while ((rxup->ub_src & TIDPVMD) || rxup->ub_cod == TC_EOC ||
		   rxup->ub_fid == tseq_get_lastrecv(rxup->ub_src) + 1);
      }
    }
#else /* ! DPVM */

    /*
     * if start of message, make new umbuf, add to frag pile
     */
    if (ff & FFSOM)
    {
      fp->fr_len -= TTMSGHDR;
      fp->fr_dat += TTMSGHDR;
      rxup = midtobuf(umbuf_new());
      rxup->ub_cod = pvmget32(cp);
      rxup->ub_enc = pvmget32(cp + 4);
      rxup->ub_wid = pvmget32(cp + 8);
      /*
       * csum = pvmget32(cp + 12);
       */
      rxup->ub_src = src;
      LISTPUTBEFORE(pcbp->tt_rxfrag, rxup, ub_link, ub_rlink);

    } else
    {
      /* locate frag's message */

      for (rxup = pcbp->tt_rxfrag->ub_link; rxup != pcbp->tt_rxfrag;
	   rxup = rxup->ub_link)
	if (rxup->ub_src == src)
	  break;
    }

    if (rxup == pcbp->tt_rxfrag)
    {				/* uh oh, no message for it */
#if 0
      for (n = 0; n < 32; n++)
	fprintf(stderr, "%x ", *(((int *) fp->fr_buf) + n));
#endif
      sprintf(pvmtxt, "mxinput() frag with no message src %x len %d ff %x\n",
	      src, len, ff);
      pvmlogerror(pvmtxt);
      fr_unref(fp);

    } else
    {
      LISTPUTBEFORE(rxup->ub_frag, fp, fr_link, fr_rlink);
      rxup->ub_len += fp->fr_len;
      /*
       * if end of message, move to pvmrxlist and count it
       */
      if (ff & FFEOM)
      {
	LISTDELETE(rxup, ub_link, ub_rlink);
	rxup->ub_codef = enctovec(rxup->ub_enc);
	LISTPUTBEFORE(pvmrxlist, rxup, ub_link, ub_rlink);
	gotem++;
	if (debugmask & TDMMESSAGE)
	{
	  sprintf(pvmtxt,
		  "mxinput() src t%x route t%x cod %d len %d\n",
		  rxup->ub_src, pcbp->tt_tid,
		  rxup->ub_cod, rxup->ub_len);
	  pvmlogerror(pvmtxt);
	}
      }
    }
#endif
  }
  return gotem;
}


/*
 * mxfer()
 * 
 * Move message frags between task and pvmd or other tasks. Returns when
 * Outgoing message (if any) fully sent and (Timed out (tmout) or At
 * least one message fully received) and No message from pvmd
 * partially received. Returns >=0 the number of complete messages
 * downloaded, or negative on error.
 */

int
mxfer(mid, tmout)
  int             mid;		/* message */
  struct timeval *tmout;	/* get at least one message */
{
  struct umbuf   *txup;		/* ogm */
  struct ttpcb   *txpcbp = 0;	/* route for ogm */
  struct frag    *txfp = 0;	/* cur ogm frag or null */
  struct timeval  tin;
  struct timeval  tnow;
  struct timeval *tvp;
  fd_set          rfds, wfds;
  int             wantmore = 1;	/* nothing rxd yet and not timed out */
  int             gotem = 0;	/* count complete msgs downloaded */
  int             ff;
  int             n;
  char           *txcp = 0;	/* point to remainder of txfp */
  int             txtogo = 0;	/* len of remainder of txfp */
  int             sbf;		/* message sent by pvmmctl() */
  struct umbuf   *up;
  struct sockaddr_in sad;
  int             s;
  struct ttpcb   *pcbp;
  int             tstk[100];	/* XXX aaigh! shouldn't be a stack */
  int             tstkp = 0;
  char            dummy[TDFRAGHDR + TTFRAGHDR + TTMSGHDR];	/* for inplace data */

#ifdef DPVM
  struct timeval  onesecond = {1L, 0L};
  struct frag    *fp;
  int             sqn;
  int             pvmmctl_scan_mode = mxfer_pvmmctl_scan_mode;
#ifdef __linux__
  struct timeval  backupTVP;    /* backup for struct pointed at by tvp */
#endif

#endif

  if (tmout)
    gettimeofday(&tin, (struct timezone *) 0);

  if (txup = midtobuf(mid))
  {
    txfp = txup->ub_frag->fr_link;
    if (!txfp->fr_buf)
    {
      txfp = fr_new(MAXHDR);
      txfp->fr_dat += MAXHDR;
      LISTPUTBEFORE(txup->ub_frag, txfp, fr_link, fr_rlink);
    }
#ifdef DPVM
    for (sqn = 0, fp = txup->ub_frag->fr_link; fp != txup->ub_frag;
	 fp = fp->fr_link)
    {
	fp->fr_seq = sqn;
	sqn += fp->fr_len;
    }
    /* We don't use message sequencing for TC_EOC messages, since they
       are supposed to be out of sequence.  No need to modify other places
       in mxfer, since TC_EOC can only originate from here.  */
    txup->ub_fid = (txup->ub_cod != TC_EOC ?
		    tseq_get_newsend(txup->ub_dst) : -1);
#endif
    if (!(txpcbp = ttpcb_find(txup->ub_dst)) || txpcbp->tt_state != TTOPEN)
      txpcbp = topvmd;
  }
  if (debugmask & TDMMESSAGE)
  {
    sprintf(pvmtxt, "mxfer() mid %d tmout ", mid);
    if (tmout)
      sprintf(pvmtxt + strlen(pvmtxt), "%d.%06d",
	      tmout->tv_sec, tmout->tv_usec);
    else
      strcat(pvmtxt, "-");
    if (txup)
      sprintf(pvmtxt + strlen(pvmtxt), " dst t%x route t%x\n",
	      txup->ub_dst, txpcbp->tt_tid);
    strcat(pvmtxt, "\n");
    pvmlogerror(pvmtxt);
  }
  /*
   * XXX we go an extra time through select because wantmore gets set
   * XXX inside loop.  no big deal.  also, could avoid calling
   * gettimeofday XXX if tmout is 0.0
   */

  while (txfp || topvmd->tt_rxf || wantmore)
  {
    if (txfp || topvmd->tt_rxf)
    {				/* gotta block */
      /*
       * must block if sending any message, or receiving from pvmd
       */
#ifdef DPVM
      tvp = &onesecond;
#else
      tvp = 0;
#endif

    } else
    {
      if (wantmore)
      {
	if (tmout)
	{			/* finite time to block */
	  gettimeofday(&tnow, (struct timezone *) 0);
	  TVXSUBY(&tnow, &tnow, &tin);
	  if (TVXLTY(tmout, &tnow))
	  {			/* time's up */
	    TVCLEAR(&tnow);
	    wantmore = 0;

	  } else
	  {
	    TVXSUBY(&tnow, tmout, &tnow);
#ifdef DPVM
	    if (TVXLTY(&onesecond, &tnow))
		tnow = onesecond;
#endif
	  }
	  tvp = &tnow;

	} else
	{			/* forever to block */
#ifdef DPVM
	  tvp = &onesecond;
#else
	  tvp = 0;
#endif
	}

      } else
      {				/* return asap */
	TVCLEAR(&tnow);
	tvp = &tnow;
      }
    }

#ifdef DPVM
reloop:
#endif

    rfds = pvmrfds;
    FD_ZERO(&wfds);
    if (txfp)
      FD_SET(txpcbp->tt_fd, &wfds);

    if (debugmask & TDMSELECT)
    {
      if (tvp)
	sprintf(pvmtxt, "mxfer() select timeout is %d.%06d\n",
		tvp->tv_sec, tvp->tv_usec);
      else
	sprintf(pvmtxt, "mxfer() select timeout is -\n");
      pvmlogerror(pvmtxt);
      print_fdset("mxfer() rfds=", pvmnfds, &rfds);
      print_fdset("mxfer() wfds=", pvmnfds, &wfds);
    }
    /*
     * ZWH: From the Linux man pages:
     *
     * On  Linux,  timeout  is  modified to reflect the amount of
     * time not slept; most other implementations do not do this.
     * This  causes  problems  both  when  Linux code which reads
     * timeout is ported to other  operating  systems,  and  when
     * code  is  ported to Linux that reuses a struct timeval for
     * multiple selects in  a  loop  without  reinitializing  it.
     * Consider timeout to be undefined after select returns.
     *
     * Therefore we need to backup tvp before invoking select and
     * restore it afterwards
     */
#ifdef __linux__
    /* Backup */
    backupTVP.tv_sec = tvp->tv_sec;
    backupTVP.tv_usec = tvp->tv_usec;
#endif __linux__
    if ((n = select(pvmnfds, &rfds, &wfds, (fd_set *) 0, tvp)) == -1
	&& errno != EINTR)
    {
      pvmlogperror("mxfer() select");
      return PvmSysErr;
    }
#ifdef __linux__
    /* Restore */
    tvp->tv_sec = backupTVP.tv_sec;
    tvp->tv_usec = backupTVP.tv_usec;
#endif __linux__
    if (debugmask & TDMSELECT)
    {
      sprintf(pvmtxt, "mxfer() select returns %d\n", n);
      pvmlogerror(pvmtxt);
    }
    if (n == -1)
      continue;
#ifdef DPVM
    {
      sigset_t        pending;

      sigpending(&pending);
      if ((sigismember(&pending, CKPTSIGNAL) &&
	  InCriticalSection == 1) || sigismember(&pending, SIGURG))
      {
	int crit = InCriticalSection;

	while (InCriticalSection > 0)
	    EndCriticalSection();
#ifndef __linux__
	/* The signal handler should be called from within EndCriticalSection
	   above.  Under Solaris it apparently is not.  */
	sigpause(0);
#endif
	while (InCriticalSection < crit)
	    StartCriticalSection();

	/* After the migration, the only open connection is to pvmd. We
	   must reinitialize txpcbp, as it may now point to free()d memory.

	   Similarly if a peer migrated: we must check if txpcbp is still
	   valid.  */
	if (txpcbp && txpcbp != topvmd)
	{
	    if (!(txpcbp = ttpcb_find(txup->ub_dst)) ||
		txpcbp->tt_state != TTOPEN)
	    {
		txpcbp = topvmd;
		/* Any fragments sent are considered lost.  We must resend the
		   whole message.  */
		if (txfp)
		{
		    txcp = 0;
		    txtogo = 0;
		    txfp = txup->ub_frag->fr_link;
		}
	    }
	}

	wantmore = 0; /* Some messages might have been read in ttpcb_flushall/
			 dpvm_oobhandler.  */
	pvmmctl_scan_mode = MXSM_ALL;

	goto reloop;
      }
    }
#endif

    /*
     * if pvmd conn has data ready, read packets
     */

    if ((FD_ISSET(topvmd->tt_fd, &rfds) && !(mxfersingle && gotem))
#ifdef DPVM
	/* Even if there is nothing to read, we might want to rescan the
	   message queue for control messages that could be received by
	   dpvm_oobhandler or ttpcb_flushall.  */
	|| pvmmctl_scan_mode == MXSM_ALL)
#endif
    {
#ifdef DPVM
      if (pvmmctl_scan_mode == MXSM_ALL)
      {
	  /* It is better to be careful with this optimisation for DPVM, since
	     in DPVM we call mxinput from 2 additional places: dpvm_oobhandler
	     and ttpcb_flushall. There is a possibility that some control
	     messages could be skipped, so during the first iteration, we scan
	     everything.  */
	  up = pvmrxlist;
	  pvmmctl_scan_mode = MXSM_NEW;
      }
      else
#endif
      up = pvmrxlist->ub_rlink;
#ifdef DPVM
      if (FD_ISSET(topvmd->tt_fd, &rfds))
#endif
      {
      if ((n = mxinput(topvmd, 0)) < 0)
      {
	if (n != -1)
	  pvmlogerror("mxfer() mxinput bad return on pvmd sock\n");
	else if (debugmask & (TDMSELECT | TDMMESSAGE | TDMPACKET))
	  pvmlogerror("mxfer() EOF on pvmd sock\n");
	return PvmSysErr;
      }
      if (gotem += n)
	wantmore = 0;
      }

      /* pass any control messages to pvmmctl() */

#ifdef DPVM
      if (pvmmctl_scan_mode != MXSM_NONE)
#endif
      while ((up = up->ub_link) != pvmrxlist)
      {
	if (up->ub_cod >= (int) TC_FIRST && up->ub_cod <= (int) TC_LAST)
	{
	  up = up->ub_rlink;
	  if ((sbf = pvmmctl(up->ub_link)) > 0)
	  {
	    if (txfp)
	      tstk[tstkp++] = sbf;
	    else
	    {
	      txup = midtobuf(sbf);
	      txfp = txup->ub_frag->fr_link;
	      txfp = txfp->fr_buf ? txfp : 0;
#ifdef DPVM
	      for (sqn = 0, fp = txup->ub_frag->fr_link; fp != txup->ub_frag;
		   fp = fp->fr_link)
	      {
		  fp->fr_seq = sqn;
		  sqn += fp->fr_len;
	      }
	      txup->ub_fid = tseq_get_newsend(txup->ub_dst);
#endif
	      if (!(txpcbp = ttpcb_find(txup->ub_dst))
		  || txpcbp->tt_state != TTOPEN)
		txpcbp = topvmd;
	    }
	  }
	}
      }
    }
    /*
     * if task conn has data ready, read packets
     */

    for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
    {
      if (pcbp->tt_state == TTOPEN && FD_ISSET(pcbp->tt_fd, &rfds))
      {
	if ((n = mxinput(pcbp, 0)) < 0)
	{
	  if (debugmask & TDMMESSAGE)
	    pvmlogerror("mxfer() marking pcb dead\n");
	  ttpcb_dead(pcbp);
	  if (pcbp == txpcbp && txfp)
	  {
	    txpcbp = topvmd;
	    txcp = 0;
	    txtogo = 0;
	    txfp = txup->ub_frag->fr_link;
	  }
	} else
	{
	  if (gotem += n)
	    wantmore = 0;
	}

      } else if (pcbp->tt_state == TTGRNWAIT
		 && FD_ISSET(pcbp->tt_fd, &rfds))
      {
	  /* !!! WARNING! Code duplication.

	     The code below has been copied to pvm_tc_conack.  If any changes
	     are made over here, corresponding changes may need to be made
	     over there.   */
	n = sizeof(sad);
	s = accept(pcbp->tt_fd, (struct sockaddr *) & sad, &n);
	if (s == -1)
	{
	  pvmlogperror("mxfer() accept");
	  ttpcb_dead(pcbp);

	} else
	{
#ifndef NOSOCKOPT
#ifndef NOUNIXDOM
	  if (!pcbp->tt_spath)
	  {			/* not unix sock */
#endif
	    static int linger[2] = {1, 60};
	    n = 1;
	    if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
			   (char *) &n, sizeof(int)) == -1)
	      pvmlogperror("mxfer() setsockopt");
	    if (setsockopt(s, SOL_SOCKET, SO_LINGER,
			   (char *) linger, sizeof(linger)) == -1)
	      pvmlogperror("mxfer() setsockopt");
#ifndef NOSOCKBUF
	    n = TTSOCKBUF;
	    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF,
			   (char *) &n, sizeof(int)) == -1
		|| setsockopt(s, SOL_SOCKET, SO_RCVBUF,
			      (char *) &n, sizeof(n)) == -1)
	      pvmlogperror("mxfer() setsockopt SO_SNDBUF");
#endif
#ifndef NOUNIXDOM
	  }
#endif
#endif				/* NOSOCKOPT */
	  if (debugmask & TDMROUTE)
	  {
	    sprintf(pvmtxt, "mxfer() accept from t%x\n",
		    pcbp->tt_tid);
	    pvmlogerror(pvmtxt);
	  }
	  (void) close(pcbp->tt_fd);
	  /*
	   * (void)dup2(s, pcbp->tt_fd); (void)close(s);
	   */
	  pvm_fd_delete(pcbp->tt_fd, 1);
	  pcbp->tt_fd = s;
	  pcbp->tt_state = TTOPEN;
#ifdef DPVM
	  {
	      /* Send the start-of-connection message.  To avoid any further
		 races and tricky cases, we send a crude message not adhearing
		 to PVM standards, but at least we send it immediately, without
		 a danger of a signal handler being invoked in between.  */
	      char c = TT_SOC;
	      if (write(pcbp->tt_fd, &c, 1) != 1)
		  ttpcb_dead(pcbp);
	  }
#endif
	  pvm_fd_add(s, 1);
	  dpvm_set(pcbp->tt_fd);
	}
      }
    }

    /*
     * if sending and socket ready to write, send packets
     */

#ifdef DPVM
    if (txfp && txpcbp->tt_migrating && txup->ub_cod != TC_EOC)
    {
	/* We are not allowed to send any more data (but the EOC marker) via
	   this socket if we got a message that the receiving task is about
	   to migrate.  */
	txpcbp = topvmd;
	txcp = 0;
	txtogo = 0;
	txfp = txup->ub_frag->fr_link;

	goto reloop;
    }
#endif

    if (txfp && txpcbp->tt_state == TTOPEN && FD_ISSET(txpcbp->tt_fd, &wfds))
    {
      if (!txtogo)
      {
	if (txfp->fr_u.dab)	/* packed data */
	  txcp = txfp->fr_dat;
	else			/* inplace data */
	  txcp = dummy + sizeof(dummy);
	txtogo = txfp->fr_len;

	/*
	 * if this is first frag, prepend message header
	 */
	ff = 0;
	if (txfp->fr_rlink == txup->ub_frag)
	{
	  if (debugmask & TDMMESSAGE)
	  {
	    sprintf(pvmtxt, "prepend tt msghdr: cod %d enc %d wid %x len %d\n",
	    txup->ub_cod, txup->ub_enc, txup->ub_wid, txup->ub_len);
	    pvmlogerror(pvmtxt);
	  }
	  txcp -= TTMSGHDR;
	  txtogo += TTMSGHDR;
	  pvmput32(txcp, txup->ub_cod);
	  pvmput32(txcp + 4,
		   (txup->ub_enc == 2 ? pvmmyndf : txup->ub_enc));
	  pvmput32(txcp + 8, txup->ub_wid);
#ifdef DPVM
	  setublen(txup);
	  pvmput32(txcp + 12, txup->ub_len);
#else
	  /*
	   * pvmput32(txcp + 12, csum);
	   */
#endif

	  ff = FFSOM;
	}
	if (txfp->fr_link == txup->ub_frag)
	  ff |= FFEOM;
#ifdef DPVM
	/*
	 * prepend tt frag header
	 */
	if (debugmask & TDMMESSAGE)
	  pvmlogerror("prepend tt fraghdr\n");
	txcp -= TTFRAGHDR;
	txtogo += TTFRAGHDR;
	pvmput32(txcp, txup->ub_fid);
	pvmput32(txcp + 4, txfp->fr_seq);
#endif
	/*
	 * prepend frag header
	 */
	if (debugmask & TDMMESSAGE)
	  pvmlogerror("prepend td fraghdr\n");
	txcp -= TDFRAGHDR;
	pvmput32(txcp, txup->ub_dst);
	pvmput32(txcp + 4, pvmmytid);
	pvmput32(txcp + 8, txtogo);
#ifdef DPVM
	pvmput32(txcp + 12, 0);
	pvmput32(txcp + 16, 0);	/* to keep putrify happy */
	pvmput8(txcp + 16, ff);
#else
	pvmput32(txcp + 12, 0);	/* to keep putrify happy */
	pvmput8(txcp + 12, ff);
#endif
	txtogo += TDFRAGHDR;
	if (!txfp->fr_u.dab)	/* inplace data */
	  txtogo -= txfp->fr_len;
      }
      if (debugmask & TDMPACKET)
      {
	sprintf(pvmtxt, "mxfer() %s dst t%x id %d seq %dn=%d\n",
		txpcbp == topvmd ?
		"[over pvmd]" :
		"[directly]",
		txup->ub_dst, txup->ub_fid,
		txfp->fr_seq, txtogo);
	pvmlogerror(pvmtxt);
      }
#if defined(IMA_RS6K) || defined(IMA_POWER4)
      n = write(txpcbp->tt_fd, txcp, min(txtogo, 4096));
#else
      n = write(txpcbp->tt_fd, txcp, txtogo);
#endif
      if (debugmask & TDMPACKET)
      {
	sprintf(pvmtxt,
		"mxfer() wrote %d\n", n);
	pvmlogerror(pvmtxt);
      }
      if (n == -1 && errno != EWOULDBLOCK && errno != EINTR)
      {
	if (txpcbp == topvmd)
	{
	  pvmlogperror("mxfer() write pvmd sock");
	  return PvmSysErr;

	} else
	{
	  pvmlogperror("mxfer() write tt sock");
	  /* reset message and route to pvmd */
	  ttpcb_dead(txpcbp);
	  txpcbp = topvmd;
	  txcp = 0;
	  txtogo = 0;
	  txfp = txup->ub_frag->fr_link;
	}
      }
      if (n > 0)
      {
	txcp += n;
	if ((txtogo -= n) <= 0)
	{
	  if (txcp == dummy + sizeof(dummy))
	  {			/* inplace */
	    txcp = txfp->fr_dat;
	    txtogo = txfp->fr_len;

	  } else
	  {			/* packed */
	    txcp = 0;
	    txfp = txfp->fr_link;
	    if (!txfp->fr_buf)
	    {
	      if (tstkp > 0)
	      {
		txup = midtobuf(tstk[--tstkp]);
		txfp = txup->ub_frag->fr_link;
		txfp = txfp->fr_buf ? txfp : 0;
#ifdef DPVM
		for (sqn = 0, fp = txup->ub_frag->fr_link; fp != txup->ub_frag;
		     fp = fp->fr_link)
		{
		    fp->fr_seq = sqn;
		    sqn += fp->fr_len;
		}
		txup->ub_fid = tseq_get_newsend(txup->ub_dst);
#endif
		if (!(txpcbp = ttpcb_find(txup->ub_dst))
		    || txpcbp->tt_state != TTOPEN)
		  txpcbp = topvmd;
	      } else
		txfp = 0;
	    }
	  }
	}
      }
    }
    if (debugmask & TDMMESSAGE)
    {
      sprintf(pvmtxt, "mxfer() txfp %d gotem %d tt_rxf %d\n",
	      (txfp ? 1 : 0), gotem, (topvmd->tt_rxf ? 1 : 0));
      pvmlogerror(pvmtxt);
    }
  }

  return gotem;
}


/*
 * msendrecv()
 * 
 * Single op to send a system message (usually to our pvmd) and get the
 * reply. Returns message handle or negative if error.
 */

int
msendrecv(other, code)
  int             other;	/* dst, src tid */
  int             code;		/* message code */
{
  int             cc;
  struct umbuf   *up;

  if (pvmsbufmid <= 0)
    return PvmNoBuf;

  /* send code to other */
  if (debugmask & TDMMESSAGE)
  {
    sprintf(pvmtxt, "msendrecv() to t%x code %d\n", other, code);
    pvmlogerror(pvmtxt);
  }
  if ((cc = mroute(pvmsbufmid, other, code, (struct timeval *) 0)) < 0)
    return cc;

  /* recv code from other */
  for (up = pvmrxlist->ub_link; 1; up = up->ub_link)
  {
    if (up == pvmrxlist)
    {
      up = up->ub_rlink;
      if ((cc = mroute(0, 0, 0, (struct timeval *) 0)) < 0)
	return cc;
      up = up->ub_link;
    }
    if (debugmask & TDMMESSAGE)
    {
      sprintf(pvmtxt, "msendrecv() from t%x code %d\n",
	      up->ub_src, up->ub_cod);
      pvmlogerror(pvmtxt);
    }
    if (up->ub_src == other && up->ub_cod == code)
      break;
  }
  LISTDELETE(up, ub_link, ub_rlink);
  if (pvmrbufmid > 0)
    umbuf_free(pvmrbufmid);
  pvmrbufmid = 0;
  if (cc = pvm_setrbuf(up->ub_mid))
    return cc;
  return up->ub_mid;
}


/*
 * mksocs()
 * 
 * Make socket for talking to pvmd.  Connect to address advertized in
 * pvmd socket-name file.
 * 
 * Returns 0 if okay, else error code.
 */

int
mksocs()
{
  char            buf[128];
  int             d;
  int             n;
  int             try;
  char           *p;

#ifndef NOUNIXDOM
  struct sockaddr_un uns;

#endif
  struct passwd  *pwd;

  if (topvmd && !pvmmigrating)
    return 0;

  /*
   * get addr of pvmd, make socket to talk. first try envar PVMSOCK,
   * then try sockaddr file.
   */

#ifndef DPVM
  if (!(p = getenv("PVMSOCK")))
  {
#endif
    if (!(p = pvmdsockfile()))
    {
      pvmlogerror("mksocs() pvmdsockfile() failed\n");
      goto bail;
    }
    if ((d = open(p, O_RDONLY, 0)) == -1)
    {
      pvmlogperror(p);
      goto bail;
    }
    try = 3;
    do
    {
      if ((n = read(d, buf, sizeof(buf))) == -1)
      {
	pvmlogperror("mksocs() read addr file");
	goto bail;
      }
      /*
       * if (n != 13) XXX danger
       */
      if (n == 0)
      {
	if (--try < 1)
	{
	  pvmlogerror("mksocs() read addr file: wrong length read\n");
	  goto bail;
	}
	sleep(1);		/* XXX hmm */
      }
    } while (n == 0);
    /*
     * while (n != 13); XXX
     */
    (void) close(d);
    buf[n] = 0;
    p = buf;
#ifndef DPVM
  }
#endif

  FD_ZERO(&pvmrfds);
  /*
   * FD_ZERO(&pvmwfds);
   */
  pvmnfds = 0;

  if (!pvmmigrating)
    topvmd = ttpcb_new();
  topvmd->tt_tid = TIDPVMD;

  if (p[0] == '/')
  {
#ifdef NOUNIXDOM
    pvmlogerror("mksocs() no support for Unix domain socket\n");
    goto bail;

#else				/* NOUNIXDOM */
    if ((topvmd->tt_fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
    {
      pvmlogperror("mksocs() socket");
      goto bail;
    }
    /*
     * XXX these 2 loops are a hack - XXX it keeps too many tasks
     * from overflowing the pvmd listen q. XXX maybe the pvmd should
     * make a socket pair for the task instead.
     */
    try = 5;
    while (1)
    {
      BZERO((char *) &uns, sizeof(uns));
      uns.sun_family = AF_UNIX;
      strcpy(uns.sun_path, p);
      n = sizeof(uns);
      if (connect(topvmd->tt_fd, (struct sockaddr *) & uns, n) == -1)
      {
	if (--try <= 0)
	{
	  pvmlogperror("mksocs() connect");
	  goto bail;
	}
	sleep(1);		/* XXX hmm again */

      } else
	break;
    }
#endif				/* NOUNIXDOM */

  } else
  {
    if ((topvmd->tt_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    {
      pvmlogperror("mksocs() socket");
      goto bail;
    }
    try = 5;
    while (1)
    {				/* goto is at the root of all good
				 * programming style */
      hex_inadport(p, &topvmd->tt_osad);
      topvmd->tt_osad.sin_family = AF_INET;
      n = sizeof(topvmd->tt_osad);

      if (connect(topvmd->tt_fd, (struct sockaddr *) & topvmd->tt_osad, n)
	  == -1)
      {
	if (--try <= 0)
	{
	  pvmlogperror("mksocs() connect");
	  goto bail;
	}
	sleep(1);		/* XXX hmm again */

      } else
	break;
    }

#ifndef NOSOCKOPT
    d = 1;
    if (setsockopt(topvmd->tt_fd, IPPROTO_TCP, TCP_NODELAY, (char *) &d, sizeof(int))
	== -1)
    {
      pvmlogperror("mksocs() setsockopt");
      goto bail;
    }
#endif
#if 0 /* DPVM - only needed for out-of-bands data.  */
    fcntl(topvmd->tt_fd, F_SETOWN, pvmmyupid);
#endif
    d = sizeof(topvmd->tt_sad);
    if (getsockname(topvmd->tt_fd, (struct sockaddr *) & topvmd->tt_sad, &d)
	== -1)
    {
      pvmlogperror("mksocs() getsockname");
      goto bail;
    }
  }

  topvmd->tt_state = TTOPEN;
  pvm_fd_add(topvmd->tt_fd, 1);

  if (!pvmmigrating)
  {
    p = getenv("DPVM_CKPTDIR");
    if (p == NULL)
    {
      pwd = getpwuid(getuid());
      if (pwd == NULL)
	strcpy(pvmckptdir, ".");
      else
      {
	p = getenv("PVM_ARCH");
	if (p == NULL)
	  strcpy(pvmckptdir, pwd->pw_dir);
	else
	  sprintf(pvmckptdir, "%s/pvm3/bin/%s",
		  pwd->pw_dir, p);
      }
    } else
    {
      strncpy(pvmckptdir, p, sizeof pvmckptdir);
      pvmckptdir[(sizeof pvmckptdir) - 1] = 0;
    }
  }
  return 0;

bail:
  if (topvmd)
    ttpcb_delete(topvmd);
  topvmd = 0;
  return PvmSysErr;
}


/*
 * unmksocs()
 * 
 * Close socket to local pvmd.
 */

int
unmksocs()
{
  if (!topvmd)
    return 1;
  while (ttlist->tt_link != ttlist)
    ttpcb_delete(ttlist->tt_link);
  if (!pvmmigrating)
  {
    ttpcb_delete(topvmd);
    topvmd = 0;
  }
  return 0;
}

#ifdef DPVM
/*
 * dpvm_oobhandler()
 *
 * Handler for the out-of-band data SIGURG signal.
 * Currently, this signal is only sent when another process is migrating.
 *
 */
void
dpvm_oobhandler(int signo)
{
    int tid = -1, n;
    int dummy, sbf, cc;
    struct ttpcb* pcbp;
    struct timeval ztv = { 0, 0 };

#ifdef __linux__
    signal(SIGURG, dpvm_oobhandler);
#endif

#if 0 /* This code should not be needed.  */
    /* See pvmmct() for an explanation why the below is necessary.  */
    if (tc_moved_handled_by_pvmmctl > 0)
    {
	if (debugmask & TDMROUTE)
	{
	    sprintf(pvmtxt, "dpvm_oobhandler(): handled by pvmmctl, count %d\n",
		    tc_moved_handled_by_pvmmctl);
	    pvmlogerror(pvmtxt);
	}
	tc_moved_handled_by_pvmmctl--;
	return;
    }
#endif

#if 0 /* We don't use out-of-band data because of its limitations.  */
    while ((n = recv(topvmd->tt_fd, &tid, sizeof(tid), MSG_OOB)) == -1 &&
	   errno == EWOULDBLOCK)
    {
	/* No OOB data yet: try to wait, read some of the ordinary data.  */
	if (debugmask & TDMMESSAGE)
	    pvmlogerror("dpvm_oobhandler(): no oob data yet\n");
	/**/ /* Not finished! */
    }

    if (n != sizeof(tid))
    {
	pvmlogerror("dpvm_oobhandler(): invalid message received\n");
	return;
    }
#else
    for (; tid == -1;)
    {
	fd_set rfds;

	FD_ZERO(&rfds);
	FD_SET(topvmd->tt_fd, &rfds);

	if ((n = select(topvmd->tt_fd + 1, &rfds, 0, 0, 0)) == -1)
	{
	    pvmlogperror("dpvm_oobhandler(): select");
	    return;
	}

	if (n > 0)
	{
	    struct umbuf* up;

	    if (mxinput(topvmd, 0) < 0)
	    {
		pvmlogerror("dpvm_oobhandler(): topvmd mxinput failed");
		return;
	    }

	    for (up = pvmrxlist->ub_link; up != pvmrxlist; up = up->ub_link)
	    {
		if (up->ub_cod == (int) TC_MOVED)
		{
		    int rbf;

		    if (debugmask & TDMROUTE)
		    {
			sprintf(pvmtxt, "dpvm_oobhandler(): received TC_MOVED\n");
			pvmlogerror(pvmtxt);
		    }
		    rbf = pvm_setrbuf(up->ub_mid);
		    pvm_upkint(&tid, 1, 1);
		    pvm_setrbuf(rbf);
		    umbuf_free(up->ub_mid);
		    break;
		}
	    }
	}
    }
#endif

    if (debugmask & TDMROUTE)
    {
	sprintf(pvmtxt, "dpvm_oobhandler(): t%x about to migrate\n", tid);
	pvmlogerror(pvmtxt);
    }

    if (!(pcbp = ttpcb_find(tid)))
    {
	if (debugmask & TDMROUTE)
	{
	    sprintf(pvmtxt, "dpvm_oobhandler(): no direct connection to t%x\n", tid);
	    pvmlogerror(pvmtxt);
	}
	return;
    }

    if (pcbp->tt_state != TTOPEN)
    {
	if (debugmask & TDMROUTE)
	{
	    sprintf(pvmtxt, "dpvm_oobhandler(): direct connection to t%x state %d\n", tid, pcbp->tt_state);
	    pvmlogerror(pvmtxt);
	}
	ttpcb_dead(pcbp);
	return;
    }

    if (debugmask & TDMROUTE)
    {
	sprintf(pvmtxt, "dpvm_oobhandler(): closing direct connection to t%x\n", tid);
	pvmlogerror(pvmtxt);
    }

    /* Read all the data from the socket up to TC_EOC.  To avoid races, we
       take into consideratio that TC_EOC might have arrived already.  */
    for (;;)
    {
	fd_set fds;
	struct umbuf* up;
	struct timeval tmout = {5, 0};
	int n;

	for (up = pvmrxlist->ub_link; up != pvmrxlist; up = up->ub_link)
	    if (up->ub_cod == TC_EOC)
		break;

	if (up != pvmrxlist)
	{
	    int rbf;

	    if (debugmask & TDMROUTE)
		pvmlogerror("dpvm_oobhandler(): received TC_EOC\n");
	    umbuf_free(up->ub_mid);
	    break;
	}

	FD_ZERO(&fds);
	FD_SET(pcbp->tt_fd, &fds);
	n = select(pcbp->tt_fd + 1, &fds, 0, 0, &tmout);
	if (debugmask & TDMSELECT)
	{
	    if (n < 0)
		pvmlogperror("dpvm_oobhandler() select");
	    sprintf(pvmtxt, "dpvm_oobhandler() select returns %d\n", n);
	    pvmlogerror(pvmtxt);
	}
	if (n > 0)
	    if (mxinput(pcbp, 0) < 0)
	    {
		pvmlogerror("dpvm_oobhandler() ttpcb mxinput failed\n");
		ttpcb_dead(pcbp);
		return;
	    }
    }

    /* We can't allow mxfer() to send any more data via this socket.  */
    pcbp->tt_migrating = 1;

    /* Prepare end of connection marker message.  */
    sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
    dummy = 0;
    pvm_pkint(&dummy, 1, 1);
    cc = mroute(pvmsbufmid, tid, TC_EOC, &ztv);
    if (cc < 0)
    {
	sprintf(pvmtxt, "dpvm_oobhandler(): mroute failed, code %d\n", cc);
	pvmlogerror(pvmtxt);
    }
    pvm_freebuf(pvm_setsbuf(sbf));
    if (cc < 0)
	return;
    if (debugmask & TDMROUTE)
	pvmlogerror("dpvm_oobhandler(): sent TC_EOC\n");

    if (!(pcbp = ttpcb_find(tid)) || pcbp->tt_state != TTOPEN)
    {
	/* It can happen that the connection is already closed in mroute()
	   above.  So don't wait if that's the case.  */
	if (debugmask & TDMROUTE)
	    pvmlogerror("dpvm_oobhandler() ttpcb connection closed in mroute\n");
	return;
    }

    /* Now we just wait for EOF.  */
    for (;;)
    {
	fd_set fds;
	struct timeval tmout = {5, 0};

	FD_ZERO(&fds);
	FD_SET(pcbp->tt_fd, &fds);
	n = select(pcbp->tt_fd + 1, &fds, 0, 0, &tmout);
	if (debugmask & TDMSELECT)
	{
	    if (n < 0)
		pvmlogperror("dpvm_oobhandler() select2");
	    sprintf(pvmtxt, "dpvm_oobhandler() select2 returns %d\n", n);
	    pvmlogerror(pvmtxt);
	}
	if (n > 0)
	{
	    if (mxinput(pcbp, 0) < 0)
	    {
		if (debugmask & TDMROUTE)
		    pvmlogerror("dpvm_oobhandler() ttpcb connection closed\n");
		ttpcb_dead(pcbp);
		break;
	    }
	    else
		pvmlogerror("dpvm_oobhandler() mxinput2 read data\n");
	}
    }
}
#endif


/*
 * pvmbeatask()
 * 
 * Initialize libpvm, config process as a task. This is called as the
 * first step of each libpvm function so no explicit initialization
 * is required.
 * 
 * Makes socket to talk to local pvmd, connects to pvmd, trades config
 * information and authentication.
 * 
 * Returns 0 if okay, else error code.
 */

int
pvmbeatask()
{
  int             sbf = 0, rbf = 0;	/* saved rx and tx message
					 * handles */
  int             prver;	/* protocol version */

#ifdef DPVM
  static int      altpid, signal_set = 0;
  extern void     dpvm_toggledebug();
  struct sigaction sa;
  char            lname[512];

#ifdef LOG_STDERR
  extern int      logfd;

#endif
#else
  static int      altpid = -1;	/* pid assigned by pvmd if diff from
				 * real */

#endif	/* DPVM */
  int             cc;
  char            authfn[LEN_OF_TMP_NAM];	/* auth file name */
  int             authfd = -1;	/* auth fd to validate pvmd ident */
  int             i;
  char            buf[16];	/* for converting sockaddr */
  char           *p;
  int             x;

  ckpt_error(ERR_DEBUG, "dpvm: in beatask");

#ifdef DPVM
  StartCriticalSection();
#endif
  /*
   * pvm_setopt(PvmDebugMask, -1);
   */
  if (pvmmytid != -1)
  {
#ifdef DPVM
    EndCriticalSection();
#endif
    return 0;
  }
#ifdef DPVM
  if (!signal_set)
  {
#ifndef __linux__
    sa.sa_handler = dpvm_toggledebug;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGUSR2, &sa, NULL);

    sa.sa_handler = dpvm_oobhandler;
    sigemptyset(&sa.sa_mask);
    sigaddset(&sa.sa_mask, CKPTSIGNAL);
    sa.sa_flags = 0;
    sigaction(SIGURG, &sa, NULL);
#else
    /* The problem with Linux that we have is that we include glibc 2 headers,
       but link with libc5, while sigset_t is not binary-identical in them.  */
    signal(SIGUSR2, dpvm_toggledebug);
    signal(SIGURG, dpvm_oobhandler);
#endif
  }
#endif
/* TG+ */
    /* TG: added signal USR1 handler installation here */
    ckpt_error(ERR_DEBUG, "dpvm: beatask is installing SIGUSR1 signal handler\n");
    fflush(stdout);
    signal(SIGUSR1, ckpt_handler);
/* TG- */


  if (x = pvmtoplvl)
    pvmtoplvl = 0;

  authfn[0] = 0;

  pvmmyupid = getpid();

  /*
   * get expected pid from environment in case we were started by the
   * pvmd and someone forked again
   */

  ckpt_error(ERR_DEBUG, "dpvm: beatask will get original pid");

#ifdef DPVM
  if (otid == -1)
  {
    if (p = getenv("PVMEPID"))
      altpid = atoi(p);
    else
      altpid = 0;
  } else
  {
    altpid = pvmmyupid;
  }
#else
  if (altpid == -1)
  {
    char           *p;

    if (p = getenv("PVMEPID"))
      altpid = atoi(p);
    else
      altpid = 0;
  }
#endif	/* DPVM */

  /*
   * if ((pvm_useruid = getuid()) == -1)
   */
  if ((pvm_useruid = geteuid()) == -1)
  {
    pvmlogerror("can't getuid()\n");
    cc = PvmSysErr;
    goto bail2;
  }
  ckpt_error(ERR_DEBUG, "dpvm: beatask will make sockets");

  if (cc = mksocs())		/* make socket to talk to pvmd */
    goto bail2;

  /*
   * initialize received-message list
   */


  if (pvmrxlist == NULL)
  {
    ckpt_error(ERR_DEBUG, "dpvm: beatask will initialize receive list");
    pvmrxlist = TALLOC(1, struct umbuf, "umb");
    BZERO((char *) pvmrxlist, sizeof(struct umbuf));
    pvmrxlist->ub_link = pvmrxlist->ub_rlink = pvmrxlist;

    ttlist = TALLOC(1, struct ttpcb, "tpcb");
    BZERO((char *) ttlist, sizeof(struct ttpcb));
    ttlist->tt_link = ttlist->tt_rlink = ttlist;
#ifdef DPVM
    tslist = TALLOC(1, struct tseq, "tseq");
    BZERO((char *) tslist, sizeof(struct tseq));
    tslist->ts_link = tslist->ts_rlink = tslist;

    oosrxlist = TALLOC(1, struct umbuf, "umb");
    BZERO((char *) oosrxlist, sizeof(struct umbuf));
    oosrxlist->ub_link = oosrxlist->ub_rlink = oosrxlist;
#endif
  }
  /*
   * create empty t-auth file that pvmd must write and we'll check
   * later
   */

  ckpt_error(ERR_DEBUG, "dpvm: beatask will create t-auth file");

  (void) TMPNAMFUN(authfn);
  if ((authfd = open(authfn, O_RDONLY | O_CREAT | O_TRUNC, 0600)) == -1)
  {
    pvmlogperror(authfn);
    pvmlogerror("pvmbeatask() can't creat t-auth file\n");
    cc = PvmSysErr;
    goto bail2;
  }
  sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
  rbf = pvm_setrbuf(0);

  /*
   * send first connect message to pvmd
   */

  ckpt_error(ERR_DEBUG, "dpvm: send first connect message");

  mxfersingle = 1;
  prver = TDPROTOCOL;
  pvm_pkint(&prver, 1, 1);
  pvm_pkstr(authfn);
  if ((cc = msendrecv(TIDPVMD, TM_CONNECT)) <= 0)
    goto bail;
  pvm_upkint(&prver, 1, 1);
  if (prver != TDPROTOCOL)
  {
    sprintf(pvmtxt, "pvmbeatask() t-d protocol mismatch (%d/%d)\n",
	    TDPROTOCOL, prver);
    pvmlogerror(pvmtxt);
    cc = PvmSysErr;
    goto bail;
  }
  pvm_upkint(&cc, 1, 1);
  if (!cc)
  {
    pvmlogerror("pvmbeatask() pvmd refuses connection\n");
    goto bail;
  }
  /*
   * check our t-auth file; write in pvmd d-auth file
   */

  ckpt_error(ERR_DEBUG, "dpvm: beatask checks t-auth file");

  if ((cc = read(authfd, (char *) &cc, 1)) == -1)
  {
    pvmlogperror("pvmbeatask() read authfile");
    cc = PvmSysErr;
    goto bail;
  }
  if (cc != 1)
  {
    pvmlogerror("pvmbeatask() pvmd didn't validate itself\n");
    cc = PvmSysErr;
    goto bail;
  }
  (void) close(authfd);
  (void) unlink(authfn);

  pvm_upkstr(authfn);
  if ((authfd = open(authfn, O_WRONLY, 0)) == -1)
  {
    pvmlogperror(authfn);
    pvmlogerror("pvmbeatask() failed to open d-auth file\n");
    authfn[0] = 0;
    cc = PvmSysErr;
    goto bail;
  }
  cc = write(authfd, authfn, 1);
  (void) close(authfd);
  if (cc != 1)
  {
    if (cc == -1)
      pvmlogperror(authfn);
    pvmlogerror("pvmbeatask() can't write d-auth file\n");
    authfn[0] = 0;
    cc = PvmSysErr;
    goto bail;
  }
  authfd = -1;
  authfn[0] = 0;

  /*
   * send second connect message to pvmd
   */

  ckpt_error(ERR_DEBUG, "dpvm: beatask send second connect message");

  pvm_initsend(PvmDataFoo);
  pvm_pkint(&pvmmyupid, 1, 1);
  pvm_pkint(&altpid, 1, 1);
#ifdef DPVM
  pvm_pkint(&otid, 1, 1);
  pvm_pkint(&optid, 1, 1);
#endif	/* DPVM */
  if ((cc = msendrecv(TIDPVMD, TM_CONN2)) <= 0)
    goto bail;
  pvm_upkint(&cc, 1, 1);
  if (!cc)
  {
    pvmlogerror("pvmbeatask() pvmd refuses connection\n");
    goto bail;
  }
  pvm_upkint(&pvmmytid, 1, 1);
  pvm_upkint(&pvmmyptid, 1, 1);

  ckpt_error(ERR_DEBUG, "dpvm: beatask resets old tid");

#ifdef DPVM
  if (otid == -1)
    otid = pvmmytid;
  if (optid == -1)
    optid = pvmmyptid;
#endif	/* DPVM */

  ckpt_error(ERR_DEBUG, "dpvm: beatask old tid is %d", otid);

  pvm_upkint(&pvmouttid, 1, 1);
  pvm_upkint(&pvmoutcod, 1, 1);
  pvmcouttid = pvmouttid;
  pvmcoutcod = pvmoutcod;
  pvm_upkint(&pvmtrctid, 1, 1);
  pvm_upkint(&pvmtrccod, 1, 1);
  pvmctrctid = pvmtrctid;
  pvmctrccod = pvmtrccod;

  ckpt_error(ERR_DEBUG, "dpvm: beatask hello 1");
  p = getenv("PVMTMASK");
  ckpt_error(ERR_DEBUG, "dpvm: beatask hello 1.1");
  if (p)
    strlen(p);
  ckpt_error(ERR_DEBUG, "dpvm: beatask hello 1.2");

  /* get trace mask from envar or zero it */

  if ((p = getenv("PVMTMASK")) && strlen(p) + 1 == TEV_MASK_LENGTH)
    BCOPY(p, pvmtrcmask, TEV_MASK_LENGTH);
  else
    TEV_INIT_MASK(pvmtrcmask);

  ckpt_error(ERR_DEBUG, "dpvm: beatask hello 2");

  BCOPY(pvmtrcmask, pvmctrcmask, TEV_MASK_LENGTH);

  ckpt_error(ERR_DEBUG, "dpvm: beatask hello 3");

  pvm_upkint(&pvmudpmtu, 1, 1);
  pvmfrgsiz = pvmudpmtu;

  pvm_upkint(&pvmmyndf, 1, 1);

  pvm_upkstr(buf);
  hex_inadport(buf, &pvmourinet);
  pvmourinet.sin_family = AF_INET;
  pvmourinet.sin_port = 0;

  pvm_upkint(&pvmschedtid, 1, 1);

  pvm_freebuf(pvm_setrbuf(rbf));
  pvm_freebuf(pvm_setsbuf(sbf));
  mxfersingle = 0;

  pvmtoplvl = x;

#if defined(DPVM) && defined(LOG_STDERR)
  if (logfd == -1 && pvmmyptid != 0)
  {
    sprintf(lname, "%s.%x", LOGFILE, pvmmytid);
    logfd = open(lname, O_WRONLY | O_CREAT | O_TRUNC, 0644);
    dup2(logfd, 2);
    printf("set fd 2 to %d\n", logfd);
  }
#endif


  ckpt_error(ERR_DEBUG, "dpvm: beatask will return");

#ifdef DPVM
#ifdef MONITOR
  mon_createthread(&__pvm_ml);
#endif
  EndCriticalSection();
#endif
  return 0;

bail:
  if (pvm_getrbuf() > 0)
    pvm_freebuf(pvm_getrbuf());
  if (pvm_getsbuf() > 0)
    pvm_freebuf(pvm_getsbuf());
  pvm_setrbuf(rbf);
  pvm_setsbuf(sbf);

  if (authfd != -1)
    (void) close(authfd);
  if (authfn[0])
    (void) unlink(authfn);
  unmksocs();

bail2:
  pvmtoplvl = x;
#ifdef DPVM
  EndCriticalSection();
#endif
  return cc;
}


int
pvmendtask()
{

#ifdef DPVM
  StartCriticalSection();
#endif
  if (pvmmytid != -1)
  {
    unmksocs();
    pvmmytid = -1;
    pvmtrctid = 0;
  }
#ifdef DPVM
  EndCriticalSection();
#endif
  return 0;
}


/************************
 **  Libpvm Functions  **
 **                    **
 ************************/


int
pvm_getopt(what)
  int             what;
{
  int             rc = 0;
  int             err = 0;
  int             x;

#ifdef DPVM
  StartCriticalSection();
#endif


  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (pvmmytid != -1 && TEV_DO_TRACE(TEV_GETOPT0))
    {
      pvm_pkint(&what, 1, 1);
      TEV_FIN;
    }
  }
  switch (what)
  {
    case PvmRoute:
      rc = pvmrouteopt;
      break;

    case PvmDebugMask:
      rc = debugmask;
      break;

    case PvmAutoErr:
      rc = pvmautoerr;
      break;

    case PvmOutputTid:
      rc = pvmcouttid;
      break;

    case PvmOutputCode:
      rc = pvmcoutcod;
      break;

    case PvmTraceTid:
      rc = pvmctrctid;
      break;

    case PvmTraceCode:
      rc = pvmctrccod;
      break;

    case PvmFragSize:
      rc = pvmfrgsiz;
      break;

    case PvmResvTids:
      rc = pvmrescode;
      break;

    case PvmSelfOutputTid:
      rc = pvmouttid;
      break;

    case PvmSelfOutputCode:
      rc = pvmoutcod;
      break;

    case PvmSelfTraceTid:
      rc = pvmtrctid;
      break;

    case PvmSelfTraceCode:
      rc = pvmtrccod;
      break;

    case PvmShowTids:
      rc = pvmshowtaskid;
      break;

    default:
      err = 1;
      break;
  }

  if (x)
  {
    if (pvmmytid != -1 && TEV_DO_TRACE(TEV_GETOPT1))
    {
      pvm_pkint(&rc, 1, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
#ifdef DPVM
  EndCriticalSection();
#endif
  if (err)
    return lpvmerr("pvm_getopt", PvmBadParam);
  return rc;
}


int
pvm_setopt(what, val)
  int             what;
  int             val;
{
  int             rc = 0;
  int             err = 0;
  int             sbf, rbf;
  int             x;
  char            buf[16];

#ifdef DPVM
  StartCriticalSection();
#endif

  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (pvmmytid != -1 && TEV_DO_TRACE(TEV_SETOPT0))
    {
      pvm_pkint(&what, 1, 1);
      pvm_pkint(&val, 1, 1);
      TEV_FIN;
    }
  }
  switch (what)
  {
    case PvmRoute:
      switch (val)
      {
	case PvmRouteDirect:
	case PvmAllowDirect:
#ifdef NO_DIRECT
	  rc = pvmrouteopt;
	  break;
#endif
	case PvmDontRoute:
	  rc = pvmrouteopt;
	  pvmrouteopt = val;
	  break;

	default:
	  rc = PvmBadParam;
	  err = 1;
	  break;
      }
      break;

    case PvmDebugMask:
      rc = debugmask;
      debugmask = val;
      break;

    case PvmAutoErr:
      rc = pvmautoerr;
      pvmautoerr = val;
      break;

    case PvmOutputTid:
      if (val && val != pvmmytid
	  && (val != pvmouttid || pvmcoutcod != pvmoutcod))
      {
	rc = PvmBadParam;
	err = 1;

      } else
      {
	rc = pvmcouttid;
	pvmcouttid = val;
      }
      break;

    case PvmOutputCode:
      if (pvmcouttid && pvmcouttid != pvmmytid && val != pvmoutcod)
      {
	rc = PvmBadParam;
	err = 1;

      } else
      {
	rc = pvmcoutcod;
	pvmcoutcod = val;
      }
      break;

    case PvmTraceTid:
      if (val && val != pvmmytid
	  && (val != pvmtrctid || pvmctrccod != pvmtrccod))
      {
	rc = PvmBadParam;
	err = 1;

      } else
      {
	rc = pvmctrctid;
	pvmctrctid = val;
      }
      break;

    case PvmTraceCode:
      if (pvmctrctid && pvmctrctid != pvmmytid && val != pvmtrccod)
      {
	rc = PvmBadParam;
	err = 1;

      } else
      {
	rc = pvmctrccod;
	pvmctrccod = val;
      }
      break;

    case PvmFragSize:
      if (val < TDFRAGHDR + TTMSGHDR + TTFRAGHDR + 4 || val > 1048576)
      {
	rc = PvmBadParam;
	err = 1;

      } else
      {
	rc = pvmfrgsiz;
	pvmfrgsiz = val;
      }
      break;

    case PvmResvTids:
      rc = pvmrescode;
      pvmrescode = val;
      break;

    case PvmSelfOutputTid:
      if (rc = BEATASK)
	err = 1;
      else
      {
	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	rbf = pvm_setrbuf(0);
	what = TS_OUTTID;
	pvm_pkint(&what, 1, 1);
	sprintf(buf, "%x", 0xffffffff & val);
	pvm_pkstr(buf);
	if ((rc = msendrecv(TIDPVMD, TM_SETOPT)) > 0)
	{
	  pvm_freebuf(pvm_setrbuf(rbf));
	  rc = pvmouttid;
	  pvmouttid = val;
	  pvmcouttid = pvmouttid;
	  pvmcoutcod = pvmoutcod;

	} else
	{
	  pvm_setrbuf(rbf);
	  err = 1;
	}
	pvm_freebuf(pvm_setsbuf(sbf));
      }
      break;

    case PvmSelfOutputCode:
      if (rc = BEATASK)
	err = 1;
      else
      {
	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	rbf = pvm_setrbuf(0);
	what = TS_OUTCOD;
	pvm_pkint(&what, 1, 1);
	sprintf(buf, "%x", 0xffffffff & val);
	pvm_pkstr(buf);
	if ((rc = msendrecv(TIDPVMD, TM_SETOPT)) > 0)
	{
	  pvm_freebuf(pvm_setrbuf(rbf));
	  rc = pvmoutcod;
	  pvmoutcod = val;
	  pvmcouttid = pvmouttid;
	  pvmcoutcod = pvmoutcod;

	} else
	{
	  pvm_setrbuf(rbf);
	  err = 1;
	}
	pvm_freebuf(pvm_setsbuf(sbf));
      }
      break;

    case PvmSelfTraceTid:
      if (rc = BEATASK)
	err = 1;
      else
      {
	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	rbf = pvm_setrbuf(0);
	what = TS_TRCTID;
	pvm_pkint(&what, 1, 1);
	sprintf(buf, "%x", 0xffffffff & val);
	pvm_pkstr(buf);
	if ((rc = msendrecv(TIDPVMD, TM_SETOPT)) > 0)
	{
	  pvm_freebuf(pvm_setrbuf(rbf));
	  rc = pvmtrctid;
	  pvmtrctid = val;
	  pvmctrctid = pvmtrctid;
	  pvmctrccod = pvmtrccod;

	} else
	{
	  pvm_setrbuf(rbf);
	  err = 1;
	}
	pvm_freebuf(pvm_setsbuf(sbf));
      }
      break;

    case PvmSelfTraceCode:
      if (rc = BEATASK)
	err = 1;
      else
      {
	sbf = pvm_setsbuf(pvm_mkbuf(PvmDataFoo));
	rbf = pvm_setrbuf(0);
	what = TS_TRCCOD;
	pvm_pkint(&what, 1, 1);
	sprintf(buf, "%x", 0xffffffff & val);
	pvm_pkstr(buf);
	if ((rc = msendrecv(TIDPVMD, TM_SETOPT)) > 0)
	{
	  pvm_freebuf(pvm_setrbuf(rbf));
	  rc = pvmtrccod;
	  pvmtrccod = val;
	  pvmctrctid = pvmtrctid;
	  pvmctrccod = pvmtrccod;

	} else
	{
	  pvm_setrbuf(rbf);
	  err = 1;
	}
	pvm_freebuf(pvm_setsbuf(sbf));
      }
      break;

    case PvmShowTids:
      rc = pvmshowtaskid;
      pvmshowtaskid = val;
      break;

    default:
      rc = PvmBadParam;
      err = 1;
      break;
  }

  if (x)
  {
    if (pvmmytid != -1 && TEV_DO_TRACE(TEV_SETOPT1))
    {
      pvm_pkint(&rc, 1, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
#ifdef DPVM
  EndCriticalSection();
#endif
  if (err)
    return lpvmerr("pvm_setopt", rc);
  return rc;
}


int
pvm_perror(s)
  char           *s;
{
#ifdef DPVM
  StartCriticalSection();
#endif
  if (pvmmytid == -1)
    fprintf(stderr, "libpvm [pid%d]: ", pvmmyupid);
  else
    fprintf(stderr, "libpvm [t%x]: ", pvmmytid);
  fprintf(stderr, "%s: %s\n",
	  (s ? s : "(null)"),
	  (pvm_errno <= 0 && pvm_errno > -pvm_nerr
	   ? pvm_errlist[-pvm_errno] : "Unknown Error"));
#ifdef DPVM
  EndCriticalSection();
#endif
  return 0;
}


int
pvm_getfds(fds)			/* XXX this function kinda sucks */
  int           **fds;		/* fd list return */
{
  static int     *fdlist = 0;
  static int      fdlen = 0;
  int             cc;
  int             nfds;
  struct ttpcb   *pcbp;
  int             x;

#ifdef DPVM
  StartCriticalSection();
#endif

  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (TEV_DO_TRACE(TEV_GETFDS0))
      TEV_FIN;
  }
  if (!(cc = BEATASK))
  {
    nfds = 1;
    for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
      if (pcbp->tt_state == TTOPEN || pcbp->tt_state == TTGRNWAIT)
	nfds++;

    if (fdlen < nfds)
    {
      fdlen = (nfds * 3) / 2 + 1;
      if (fdlist)
	fdlist = TREALLOC(fdlist, fdlen, int);
      else
	fdlist = TALLOC(fdlen, int, "fdlist");
    }
    fdlist[0] = topvmd->tt_fd;
    nfds = 1;
    for (pcbp = ttlist->tt_link; pcbp != ttlist; pcbp = pcbp->tt_link)
      if (pcbp->tt_state == TTOPEN || pcbp->tt_state == TTGRNWAIT)
	fdlist[nfds++] = pcbp->tt_fd;
    *fds = fdlist;
    cc = nfds;
  }
  if (x)
  {
    if (TEV_DO_TRACE(TEV_GETFDS1))
    {
      pvm_pkint(&cc, 1, 1);
      if (cc > 0)
	pvm_pkint((int *) fdlist, nfds, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
#ifdef DPVM
  EndCriticalSection();
#endif
  return (cc < 0 ? lpvmerr("pvm_getfds", cc) : cc);
}


int
pvm_start_pvmd(argc, argv, block)
  int             argc;		/* number of args to pass to pvmd (>=
				 * 0) */
  char          **argv;		/* args for pvmd or null */
  int             block;	/* if true, don't return until add
				 * hosts are started */
{
  char           *sfn;
  struct stat     sb;
  int             cc;
  char           *fn;		/* file to exec */
  char          **av;
  int             x;

#ifdef DPVM
  StartCriticalSection();
#endif
  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (pvmmytid != -1 && TEV_DO_TRACE(TEV_START_PVMD0))
    {
      pvm_pkint(&argc, 1, 1);
      pvm_pkint(&block, 1, 1);
      for (cc = 0; cc < argc; cc++)
	pvm_pkstr(argv[cc]);
      TEV_FIN;
    }
  }
  if (argc < 0 || !argv)
    argc = 0;

  if ((pvm_useruid = getuid()) == -1)
  {
    pvmlogerror("can't getuid()\n");
    cc = PvmSysErr;
    goto bail;
  }
  if (!(sfn = pvmdsockfile()))
  {
    pvmlogerror("pvm_start_pvmd() pvmdsockfile() failed\n");
    cc = PvmSysErr;
    goto bail;
  }
  if (stat(sfn, &sb) != -1)
  {
    cc = PvmDupHost;
    goto bail;
  }
  fn = pvmgetpvmd();

  av = TALLOC(argc + 2, char *, "argv");
  if (argc > 0)
    BCOPY((char *) &argv[0], (char *) &av[1], argc * sizeof(char *));
  av[0] = fn;
  av[argc + 1] = 0;

  if (!fork())
  {
    if (!fork())
    {
      for (cc = getdtablesize(); cc-- > 0;)
	(void) close(cc);
      (void) open("/dev/null", O_RDONLY, 0);
      (void) open("/dev/null", O_WRONLY, 0);
      (void) dup2(1, 2);
      execvp(av[0], av);
    }
    _exit(0);
  }
  (void) wait(0);

  PVM_FREE(av);

  for (cc = 8; cc > 0 && stat(sfn, &sb) == -1; cc--)
    sleep(1);

  if (cc <= 0)
  {
    cc = PvmCantStart;
    goto bail;
  }
  if (cc = BEATASK)
    goto bail;

  if (block)
  {
    struct pvmhostinfo *hip;
    int             t = 1;

    pvm_config((int *) 0, (int *) 0, &hip);
    while ((cc = pvm_addhosts(&hip[0].hi_name, 1, (int *) 0)) == PvmAlready)
    {
      sleep(t);
      if (t < 8)
	t *= 2;
    }
    if (cc != PvmDupHost)
      goto bail;
    cc = BEATASK;
  }
bail:

  if (x)
  {
    if (TEV_DO_TRACE(TEV_START_PVMD1))
    {
      pvm_pkint(&cc, 1, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
#ifdef DPVM
  EndCriticalSection();
#endif
  return (cc < 0 ? lpvmerr("pvm_start_pvmd", cc) : 0);
}


/*
 * tev_begin()
 * 
 * Start a trace event.  Initialize a message, write time and event
 * kind.
 */

int
tev_begin(kind)
  int             kind;
{
  struct timeval  now;

  gettimeofday(&now, (struct timezone *) 0);
  pvmtrcmid = pvm_mkbuf(PvmDataFoo);
  pvmtrcsbf = pvm_setsbuf(pvmtrcmid);
  pvm_pkint((int *) &now.tv_sec, 1, 1);
  pvm_pkint((int *) &now.tv_usec, 1, 1);
  pvm_pkint(&pvmmytid, 1, 1);
  pvm_pkint(&kind, 1, 1);
  /*
   * fprintf(stderr, "tev_begin() mid %d\n", pvmtrcmid);
   */
  return 1;
}


/*
 * tev_fin()
 * 
 * End a trace event.  Send message.
 */

int
tev_fin()
{
  int             routetmp;
  static struct timeval ztv = {0, 0};

  /*
   * fprintf(stderr, "tev_fin() mid %d\n", pvmtrcmid);
   */
  if (pvmmytid != pvmtrctid)
  {
    if ((routetmp = pvmrouteopt) == PvmRouteDirect)
      pvmrouteopt = PvmAllowDirect;
    mroute(pvmtrcmid, pvmtrctid, pvmtrccod, &ztv);
    pvmrouteopt = routetmp;
  }
  pvm_setsbuf(pvmtrcsbf);
  pvmtrcsbf = 0;
  pvm_freebuf(pvmtrcmid);
  pvmtrcmid = 0;
  return 0;
}


/*
 * tev_do_trace()
 * 
 * Export TEV_DO_TRACE() so user code doesn't have to import masks and
 * trace tid, etc.
 */

int
tev_do_trace(kind)
  int             kind;
{
  return TEV_DO_TRACE(kind);
}


int
pvm_precv(tid, tag, cp, len, dt, rtid, rtag, rlen)
  int             tid;
  int             tag;
  void           *cp;
  int             len;
  int             dt;
  int            *rtid;
  int            *rtag;
  int            *rlen;
{
  int             rbf;
  int             cc = 0;
  int             l;
  int             x;
  long            ad;

#ifdef DPVM
  StartCriticalSection();
#endif
  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (TEV_DO_TRACE(TEV_PRECV0))
    {
      pvm_pkint(&tid, 1, 1);
      pvm_pkint(&tag, 1, 1);
      ad = (long) cp;
      pvm_pklong(&ad, 1, 1);
      pvm_pkint(&len, 1, 1);
      pvm_pkint(&dt, 1, 1);
      TEV_FIN;
    }
  }
  switch (dt)
  {

    case PVM_BYTE:
      len *= sizeof(char);
      break;

    case PVM_SHORT:
    case PVM_USHORT:
      len *= sizeof(short);
      break;

    case PVM_INT:
    case PVM_UINT:
      len *= sizeof(int);
      break;

    case PVM_LONG:
    case PVM_ULONG:
      len *= sizeof(long);
      break;

    case PVM_FLOAT:
      len *= sizeof(float);
      break;

    case PVM_CPLX:
      len *= sizeof(float) * 2;
      break;

    case PVM_DOUBLE:
      len *= sizeof(double);
      break;

    case PVM_DCPLX:
      len *= sizeof(double) * 2;
      break;

    default:
      cc = PvmBadParam;
      break;
  }

  if (!cc)
  {
    rbf = pvm_setrbuf(0);
    cc = pvm_recv(tid, tag);
    if (cc > 0)
    {
      pvm_bufinfo(cc, &l, rtag, rtid);
      if (rlen)
	*rlen = l;
      if (l < len)
	len = l;
      pvm_upkbyte((char *) cp, len, 1);
      pvm_freebuf(cc);
      cc = 0;
    }
    pvm_setrbuf(rbf);
  }
  if (x)
  {
    if (TEV_DO_TRACE(TEV_PRECV1))
    {
      pvm_pkint(&cc, 1, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
  if (cc < 0)
    lpvmerr("pvm_precv", cc);
#ifdef DPVM
  EndCriticalSection();
#endif
  return cc;
}


int
pvm_psend(tid, tag, cp, len, dt)
  int             tid;
  int             tag;
  void           *cp;
  int             len;
  int             dt;
{
  int             sbf;
  int             cc = 0;
  int             x;
  long            ad;

#ifdef DPVM
  StartCriticalSection();
#endif
  if (x = pvmtoplvl)
  {
    pvmtoplvl = 0;
    if (TEV_DO_TRACE(TEV_PSEND0))
    {
      pvm_pkint(&tid, 1, 1);
      pvm_pkint(&tag, 1, 1);
      ad = (long) cp;
      pvm_pklong(&ad, 1, 1);
      pvm_pkint(&len, 1, 1);
      pvm_pkint(&dt, 1, 1);
      TEV_FIN;
    }
  }
  switch (dt)
  {

    case PVM_BYTE:
      len *= sizeof(char);
      break;

    case PVM_SHORT:
    case PVM_USHORT:
      len *= sizeof(short);
      break;

    case PVM_INT:
    case PVM_UINT:
      len *= sizeof(int);
      break;

    case PVM_LONG:
    case PVM_ULONG:
      len *= sizeof(long);
      break;

    case PVM_FLOAT:
      len *= sizeof(float);
      break;

    case PVM_CPLX:
      len *= sizeof(float) * 2;
      break;

    case PVM_DOUBLE:
      len *= sizeof(double);
      break;

    case PVM_DCPLX:
      len *= sizeof(double) * 2;
      break;

    default:
      cc = PvmBadParam;
      break;
  }

  if (!cc)
  {
    sbf = pvm_setsbuf(pvm_mkbuf(PvmDataInPlace));
    pvm_pkbyte((char *) cp, len, 1);
    if ((cc = pvm_send(tid, tag)) > 0)
      cc = 0;
    pvm_freebuf(pvm_setsbuf(sbf));
  }
  if (x)
  {
    if (TEV_DO_TRACE(TEV_PSEND1))
    {
      pvm_pkint(&cc, 1, 1);
      TEV_FIN;
    }
    pvmtoplvl = x;
  }
  if (cc < 0)
    lpvmerr("pvm_psend", cc);
#ifdef DPVM
  EndCriticalSection();
#endif
  return cc;
}

#ifdef DPVM

#ifndef BERRY
extern int      sigprocmaskharderr;
int             sigprocmaskharderr;
extern int      setsighard();

#endif

void 
StartCriticalSection(void)
{
  sigset_t        mask;

  if (InCriticalSection++ == 0)
  {
#if defined (BERRY) || defined (__linux__)
  sigemptyset(&mask);
  sigaddset(&mask, CKPTSIGNAL);
  sigaddset(&mask, SIGURG);
  sigprocmask(SIG_BLOCK, &mask, (sigset_t *) NULL);
#else
#define sigmask(n) ((unsigned int)1 << (((n) - 1) & (32 - 1)))
#define sigword(n) (((unsigned int)((n) - 1))>>5)
  /* make an educated guess what sigemptyset will do: */
  mask.__sigbits[0] = 0;
  mask.__sigbits[1] = 0;
  mask.__sigbits[2] = 0;
  mask.__sigbits[3] = 0;
  mask.__sigbits[sigword(CKPTSIGNAL)] |= sigmask(CKPTSIGNAL);
  mask.__sigbits[sigword(SIGURG)] |= sigmask(SIGURG);
  sigprocmaskhard(SIG_BLOCK, &mask, (sigset_t *) NULL);
#endif
  }
}

void 
EndCriticalSection(void)
{
  sigset_t        mask;

  if (--InCriticalSection == 0)
  {
    sigemptyset(&mask);
    sigaddset(&mask, CKPTSIGNAL);
    sigaddset(&mask, SIGURG);
#if defined (BERRY) || defined (__linux__)
    sigprocmask(SIG_UNBLOCK, &mask, (sigset_t *) NULL);
#else
    sigprocmaskhard(SIG_UNBLOCK, &mask, (sigset_t *) NULL);
#endif
  }
}

#endif	/* DPVM */


int             dpvm_txfd = -1;
static int      dpvm_max = 0;
static fd_set   dpvm_fdset;

void            dpvm_set();
void            dpvm_clr();

#if 0
void
dpvm_sigurg(sig, code)
  int             sig;
  int             code;
{
  struct fd_set   urg_set;
  struct timeval  tv;
  int             i;
  char            oob;


  urg_set = dpvm_fdset;
  TVCLEAR(&tv);

  while (1)
  {
    urg_set = dpvm_fdset;

    if ((i = select(dpvm_max, NULL, NULL, &urg_set, &tv)) == -1 &&
	errno == EINTR)
      continue;

    if (i == 0)
      break;

    for (i = 0; i < dpvm_max; i++)
    {
      if (!FD_ISSET(i, &urg_set))
	continue;

      while (recv(i, &oob, 1, MSG_OOB) == 0);
      /*
       * pvmlogerror("caught ttpcb_aclwait()\n");
       */
      if (i == dpvm_txfd)
	ttpcb_aclwait(i, 0);
      else
	ttpcb_aclwait(i, 1);

      dpvm_clr(i);
    }
  }
}

#endif


void
dpvm_set(fd)
  int             fd;
{
  static int      pid;

  pid = getpid();

  FD_SET(fd, &dpvm_fdset);
  if (fd >= dpvm_max)
    dpvm_max = fd + 1;
}


void
dpvm_clr(fd)
  int             fd;
{
  int             i;

  FD_CLR(fd, &dpvm_fdset);
  i = dpvm_max;
  while (--i, i >= 0 && !FD_ISSET(i, &dpvm_fdset));
  dpvm_max = i + 1;
}


void
dpvm_clrall()
{
  FD_ZERO(&dpvm_fdset);
  dpvm_max = 0;
}

#ifdef DPVM
static struct tseq*
tseq_new(void)
{
    struct tseq* ts;

    if (ts = TALLOC(1, struct tseq, "tseq"))
	BZERO((char*)ts, sizeof(*ts));

    return ts;
}

int
tseq_get_newsend(int tid)
{
    struct tseq* ts;

    for (ts = tslist->ts_link; ts != tslist; ts = ts->ts_link)
	if (ts->ts_tid >= tid)
	    break;

    if (ts->ts_tid != tid)
    {
	struct tseq* ts2;

	ts2 = tseq_new();
	ts2->ts_tid = tid;

	LISTPUTBEFORE(ts, ts2, ts_link, ts_rlink);

	ts = ts2;
    }

    return ++ts->ts_lastsend;
}

int
tseq_get_lastrecv(int tid)
{
    struct tseq* ts;

    for (ts = tslist->ts_link; ts != tslist; ts = ts->ts_link)
	if (ts->ts_tid >= tid)
	    break;

    return (ts->ts_tid == tid ? ts->ts_lastrecv : 0);
}

void
tseq_increase_lastrecv(int tid)
{
    struct tseq* ts;

    for (ts = tslist->ts_link; ts != tslist; ts = ts->ts_link)
	if (ts->ts_tid >= tid)
	    break;

    if (ts->ts_tid != tid)
    {
	struct tseq* ts2;

	ts2 = tseq_new();
	ts2->ts_tid = tid;

	LISTPUTBEFORE(ts, ts2, ts_link, ts_rlink);

	ts = ts2;
    }

    ts->ts_lastrecv++;
}
#endif
