#ifdef MONITOR
#include <pthread.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

#include <rpc/types.h>
#include <rpc/xdr.h>

#include <string.h>

#include <errno.h>

#include "monlib.h"

struct mon_list __pvm_ml;

sem_t
sem_create(val)
	int val;
{
	sem_t sem;

	sem = (sem_t)calloc(1, sizeof *sem);
	pthread_mutex_init(&sem->lock, NULL);
	pthread_cond_init(&sem->cond, NULL);
	sem->count = val;
	return sem;
}

int
sem_init(sem, val)
	sem_t *sem;
	int val;
{
	*sem = sem_create(val);
	return 0;
}

void
sem_wait(s)
	sem_t *s;
{
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	if (sem->count > 0) {
		sem->count--;
		pthread_mutex_unlock(&sem->lock);
		return;
	}

	pthread_cond_wait(&sem->cond, &sem->lock);
	sem->count--;
	pthread_mutex_unlock(&sem->lock);
}

void
sem_signal(s)
	sem_t *s;
{
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	sem->count++;
	pthread_cond_signal(&sem->cond);
	pthread_mutex_unlock(&sem->lock);
}

int
sem_testwait(s)
	sem_t *s;
{
	int retval;
	sem_t sem = *s;

	pthread_mutex_lock(&sem->lock);
	if (sem->count == 0) {
		pthread_mutex_unlock(&sem->lock);
		return 0;
	}
	sem->count--;
	pthread_mutex_unlock(&sem->lock);
	return 1;
}

int
sem_destroy(s)
	sem_t *s;
{
	sem_t sem = *s;

	pthread_cond_destroy(&sem->cond);
	pthread_mutex_destroy(&sem->lock);
	free(sem);
	return 0;
}

void
mon_add_data(mp, type, code, src, dst, len, stime, etime)
	struct mon_list *mp;
	int type, src, dst, len;
	struct timespec *stime, *etime;
{
	struct mon_data md;

	md.mon_montype = type;
	md.mon_msgtype = code;
	md.mon_src = src;
	md.mon_dst = dst;
	md.mon_len = len;
	md.mon_stime = *stime;
	md.mon_etime = *etime;
	monlist_addtail(mp, &md);
	sem_signal(&mp->avail);
}

int
mon_get_data(mp, mdp)
	struct mon_list *mp;
	struct mon_data *mdp;
{
	struct mon_data md;

	sem_wait(&mp->avail);
	if (mp->head == NULL && mp->flags & MON_EXITING)
		return 0;
	monlist_remhead(mp, mdp);
	return 1;
}

struct mon_list *
monlist_create()
{
	struct mon_list *mp;

	mp = (struct mon_list *)malloc(sizeof (struct mon_list));
	monlist_init(mp);
	return mp;
}

void
monlist_destroy(mp)
	struct mon_list *mp;
{
	sem_destroy(&mp->avail);
	pthread_mutex_destroy(&mp->lock);
}

void
monlist_init(mp)
	struct mon_list *mp;
{
	mp->head = mp->tail = NULL;
	sem_init(&mp->avail, 0);
	pthread_mutex_init(&mp->lock, NULL);
	mp->flags = 0;
}

void
monlist_addtail(mp, dp)
	struct mon_list *mp;
	struct mon_data *dp;
{
	struct mon_data *tp;

	pthread_mutex_lock(&mp->lock);
	tp = (struct mon_data *)malloc(sizeof (struct mon_data));
	*tp = *dp;
	tp->mon_next = 0;
	if (mp->tail == NULL)
		mp->head = mp->tail = tp;
	else {
		mp->tail->mon_next = tp;
		mp->tail = tp;
	}
	pthread_mutex_unlock(&mp->lock);
}

void
monlist_exit(mp)
	struct mon_list *mp;
{
	pthread_mutex_lock(&mp->lock);
	mp->flags |= MON_EXITING;
	pthread_mutex_unlock(&mp->lock);
	sem_signal(&mp->avail);
}

void
monlist_remhead(mp, dp)
	struct mon_list *mp;
	struct mon_data *dp;
{
	pthread_mutex_lock(&mp->lock);
	*dp = *mp->head;
	free(mp->head);
	if (mp->tail == mp->head)
		mp->tail = NULL;
	mp->head = dp->mon_next;
	pthread_mutex_unlock(&mp->lock);
}

int
xdr_timespec(xdrs, tsp)
	XDR *xdrs;
	struct timespec *tsp;
{
	if (!xdr_long(xdrs, &tsp->tv_sec))
		return FALSE;
	if (!xdr_long(xdrs, &tsp->tv_nsec))
		return FALSE;
	return TRUE;
}

int
xdr_mondata(xdrs, mdp)
	XDR *xdrs;
	struct mon_data *mdp;
{
	if (!xdr_int(xdrs, &mdp->mon_montype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_msgtype))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_src))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_dst))
		return FALSE;
	if (!xdr_int(xdrs, &mdp->mon_len))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_stime))
		return FALSE;
	if (!xdr_timespec(xdrs, &mdp->mon_etime))
		return FALSE;
	return TRUE;
}

int
mon_connect(host, port)
	char *host;
	int port;
{
	struct sockaddr_in sin;
	int s, e, family;
	struct hostent *hp, hent;
	struct in_addr saddr;
	char buffer[MAXHOSTNAMELEN+1];
	int herr;

#if 0
	if (inet_aton(host, &saddr) == 0) {
#endif
		hp = gethostbyname_r(host, &hent, buffer, sizeof buffer, &herr);
		if (hp == 0) {
			(void)fprintf(stderr, "sync: unknown host: %s\n", host);
			return -1;
		}
		family = hp->h_addrtype;
		bcopy(hp->h_addr, (char *)&saddr, hp->h_length);
#if 0
		host = hp->h_name;
	} else
		family = AF_INET;
#endif

	do {
		const int one = 1;

		/*
		 * Wonder why the socket must be recreated each time,
		 * but without this it doesn't work on AIX.
		 */
		if ((s = socket(family, SOCK_STREAM, 0)) < 0) {
			perror("client socket");
			return -1;
		}
		setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *)&one, sizeof (int));
		sin.sin_addr = saddr;
		sin.sin_port = port;
#ifdef HAS_SINLEN
		sin.sin_len = sizeof (struct sockaddr_in);
#endif
		sin.sin_family = family;
		e = connect(s, (struct sockaddr *)&sin, sizeof(sin));
		if (e < 0)
			close(s);
	} while (e < 0 && (errno == ECONNREFUSED || errno == EINVAL || errno == EADDRINUSE));
		
	if (e < 0) {
		perror("connect");
		close(s);
		return -1;
	}

	return s;
}

/*
 * XXX no error checking done in this function.
 */
int
mon_getmonaddr(hostp, portp)
	char **hostp;
	int *portp;
{
	char *host, *p;

	host = getenv("DYNAMITE_MONITOR");
	p = getenv("DYNAMITE_MONPORT");
	if (host == NULL || p == NULL)
		return -1;
	*portp = atoi(p);
	*hostp = host;
	return 0;
}

void *
mon(v)
	void *v;
{
	struct mon_list *mp = v;
	int src, dst, len, type;
	struct timespec stime, etime;
	int s, n;	
	char *host, *buf;
	int port;
	XDR xdr;
	struct mon_data md;

	buf = malloc(sizeof (struct mon_data));
	if (mon_getmonaddr(&host, &port) < 0)
		return NULL;
retry:
	s = mon_connect(host, port);
	if (s < 0)
		return NULL;

	while (mon_get_data(mp, &md)) {
#if 0
		printf("%d -> %d type %d len %d start %ld.%ld end %ld.%ld\n",
			md.mon_src, md.mon_dst, md.mon_msgtype, md.mon_len,
			md.mon_stime.tv_sec, md.mon_stime.tv_nsec,
			md.mon_etime.tv_sec, md.mon_etime.tv_nsec);
#endif
		xdrmem_create(&xdr, buf, sizeof (struct mon_data), XDR_ENCODE);
		xdr_mondata(&xdr, &md);
		xdr_destroy(&xdr);
		n = write(s, buf, sizeof (struct mon_data));
		if (n < 0 && errno == EPIPE)
			goto retry;
			
	}
	monlist_destroy(mp);
#if 0
	printf("monitor thread done.\n");
#endif
}

int
mon_createthread(mlp)
	struct mon_list *mlp;
{
	pthread_attr_t attr;

	monlist_init(mlp);
	pthread_attr_init(&attr);
	if (pthread_create(&mlp->thread, &attr, mon, mlp) < 0)
		return -1;
	return 0;
}

void
mon_terminatethread(mlp)
	struct mon_list *mlp;
{
	int n;
	pthread_t t;

	t = mlp->thread;
	monlist_exit(mlp);
	if (pthread_join(t, (void **)&n) < 0)
		perror("pthread_join");
}
#endif
